
internal void
transformVertices( __m128 * transform, __m128 * translate, vec3 * vertex_input, vec3 * vertex_output, int32 min_index, int32 max_index ) {
	for( int32 vertex_index = min_index; vertex_index < max_index; vertex_index++ ) {
		vec3 point = vertex_input[ vertex_index ];
		
		__m128 pointX = _mm_set_ps1( point.x );
		__m128 pointY = _mm_set_ps1( point.y );
		__m128 pointZ = _mm_set_ps1( point.z );
		
		__m128 posX = _mm_add_ps( _mm_add_ps( _mm_mul_ps( transform[ 0 ], pointX ), _mm_mul_ps( transform[ 3 ], pointY ) ), _mm_mul_ps( transform[ 6 ], pointZ ) );
		__m128 posY = _mm_add_ps( _mm_add_ps( _mm_mul_ps( transform[ 1 ], pointX ), _mm_mul_ps( transform[ 4 ], pointY ) ), _mm_mul_ps( transform[ 7 ], pointZ ) );
		__m128 posZ = _mm_add_ps( _mm_add_ps( _mm_mul_ps( transform[ 2 ], pointX ), _mm_mul_ps( transform[ 5 ], pointY ) ), _mm_mul_ps( transform[ 8 ], pointZ ) );
		
		__m128 traX = _mm_add_ps( posX, translate[ 0 ] );
		__m128 traY = _mm_add_ps( posY, translate[ 1 ] );
		__m128 traZ = _mm_add_ps( posZ, translate[ 2 ] );
		
		vertex_output[ vertex_index       ].x = ( ( flo32 * )&traX )[ 0 ];
		vertex_output[ vertex_index       ].y = ( ( flo32 * )&traY )[ 0 ];
		vertex_output[ vertex_index       ].z = ( ( flo32 * )&traZ )[ 0 ];
		vertex_output[ vertex_index +  72 ].x = ( ( flo32 * )&traX )[ 1 ];
		vertex_output[ vertex_index +  72 ].y = ( ( flo32 * )&traY )[ 1 ];
		vertex_output[ vertex_index +  72 ].z = ( ( flo32 * )&traZ )[ 1 ];
		vertex_output[ vertex_index + 144 ].x = ( ( flo32 * )&traX )[ 2 ];
		vertex_output[ vertex_index + 144 ].y = ( ( flo32 * )&traY )[ 2 ];
		vertex_output[ vertex_index + 144 ].z = ( ( flo32 * )&traZ )[ 2 ];
		vertex_output[ vertex_index + 216 ].x = ( ( flo32 * )&traX )[ 3 ];
		vertex_output[ vertex_index + 216 ].y = ( ( flo32 * )&traY )[ 3 ];
		vertex_output[ vertex_index + 216 ].z = ( ( flo32 * )&traZ )[ 3 ];
	}
}

internal void
animateFans( RENDER_SYSTEM_3D * renderSystem, APP_STATE * appState, MEMORY * tempMemory ) {
	PROFILE_FUNCTION();
	int32 fan_count[] = { 480, 196, 252, 196, 480, 196, 252, 196 };
	int32 counter = 0;
	
	int32 min_index[ 8 ] = {};
	int32 max_index[ 8 ] = {};
	for( int32 iter = 0; iter < 8; iter++ ) {
		min_index[ iter ] = counter;
		counter += fan_count[ iter ];
		max_index[ iter ] = counter;
	}
	
	VERTEX_BUFFER * vertexBuffer = renderSystem->vertexBuffer;
	
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, assetID_texture_stadiumFan ).bufferID;
	RENDER_MESH        mesh = getMesh(    renderSystem->assetSystem, assetID_mesh_stadiumFan );
	
	int32  position_count =  72;
	int32      face_count = 336;
	vec3 * position       = _pushArray_clear( tempMemory, vec3, position_count * 4 );
	
	uint32 vertex_index = getTempIndex( vertexBuffer );
	 int32 vertex_total = face_count * 2248;
	Assert( ( vertexBuffer->temp_count + vertex_total ) < vertexBuffer->temp_maxCount );
	
	VERTEX_TEXMESH_3D * vertex0 = vertexBuffer->buffer_base + ( vertexBuffer->temp_select * vertexBuffer->temp_maxCount );
	
	vec3 scale0 = Vec3( 1.0f, 1.0f, 1.0f );
	
	__m128 zero    = _mm_set_ps1( 0.0f  );
	__m128 point5  = _mm_set_ps1( 0.5f  );
	__m128 point59 = _mm_set_ps1( 0.59f );
	__m128  one    = _mm_set_ps1( 1.0f  );
	
	__m128 bone0[] = { zero, _mm_set_ps1( 0.5f ), zero };
	vec3 bone1A = Vec3( 0.0f, 1.0f,  0.0f );
	vec3 bone2A = Vec3( 0.0f, 1.59f, 0.0f );
	
	flo32 rangeA = -PI * 0.5f;
	flo32 rangeB = -PI * 0.8f;
	
	for( int32 loop = 0; loop < 8; loop++ ) {
		flo32 radians0 = appState->stadiumFan_rotation[ loop ];
		
		flo32 co0 = cosf( radians0 );
		flo32 si0 = sinf( radians0 );
		
		__m128  cos0 = _mm_set_ps1(  co0 );
		__m128  sin0 = _mm_set_ps1(  si0 );
		__m128 _sin0 = _mm_set_ps1( -si0 );
		
		__m128 rotation0[] = { cos0, zero, _sin0,   zero, one, zero,   sin0, zero, cos0 };
		
		for( int32 seat_index = min_index[ loop ]; seat_index < max_index[ loop ]; seat_index += 4 ) {
			flo32 t[] = {
				appState->stadiumFan_t[ seat_index + 0 ],
				appState->stadiumFan_t[ seat_index + 1 ],
				appState->stadiumFan_t[ seat_index + 2 ],
				appState->stadiumFan_t[ seat_index + 3 ],
			};
			
			flo32 radiansA[] = { rangeA *          t[ 0 ],   rangeA *          t[ 1 ],   rangeA *          t[ 2 ],   rangeA *          t[ 3 ]   };
			flo32 radiansB[] = { rangeB * ( 1.0f - t[ 0 ] ), rangeB * ( 1.0f - t[ 1 ] ), rangeB * ( 1.0f - t[ 2 ] ), rangeB * ( 1.0f - t[ 3 ] ) };
			
			flo32 coA[] = { cosf( radiansA[ 0 ] ), cosf( radiansA[ 1 ] ), cosf( radiansA[ 2 ] ), cosf( radiansA[ 3 ] ) };
			flo32 siA[] = { sinf( radiansA[ 0 ] ), sinf( radiansA[ 1 ] ), sinf( radiansA[ 2 ] ), sinf( radiansA[ 3 ] ) };
			flo32 coC[] = { cosf( radiansB[ 0 ] ), cosf( radiansB[ 1 ] ), cosf( radiansB[ 2 ] ), cosf( radiansB[ 3 ] ) };
			flo32 siC[] = { sinf( radiansB[ 0 ] ), sinf( radiansB[ 1 ] ), sinf( radiansB[ 2 ] ), sinf( radiansB[ 3 ] ) };
			
			__m128  cosA = _mm_set_ps(  coA[ 3 ],  coA[ 2 ],  coA[ 1 ],  coA[ 0 ] );
			__m128  sinA = _mm_set_ps(  siA[ 3 ],  siA[ 2 ],  siA[ 1 ],  siA[ 0 ] );
			__m128 _sinA = _mm_set_ps( -siA[ 3 ], -siA[ 2 ], -siA[ 1 ], -siA[ 0 ] );
			__m128  cosC = _mm_set_ps(  coC[ 3 ],  coC[ 2 ],  coC[ 1 ],  coC[ 0 ] );
			__m128  sinC = _mm_set_ps(  siC[ 3 ],  siC[ 2 ],  siC[ 1 ],  siC[ 0 ] );
			
			__m128 transformB_t = _mm_add_ps( _mm_mul_ps( sinA, sinA ), _mm_mul_ps( cosA, cosA ) );
			
			__m128  transformC_t = _mm_mul_ps( cosC, transformB_t );
			__m128  transformC_s = _mm_mul_ps( sinC, transformB_t );
			__m128 _transformC_s = _mm_sub_ps( zero, transformC_s );
			
			__m128 transformD[] = { cos0, zero, _sin0, _mm_mul_ps( sin0, sinA ), cosA, _mm_mul_ps( cos0, sinA ), _mm_mul_ps( sin0, cosA ), _sinA, _mm_mul_ps( cos0, cosA ) };
			__m128 transformE[] = { cos0, zero, _sin0,  zero, transformB_t, zero, _mm_mul_ps( sin0, transformB_t ), zero, _mm_mul_ps( cos0, transformB_t ) };
			__m128 transformF[] = { cos0, zero, _sin0, _mm_mul_ps( sin0, transformC_s ), transformC_t, _mm_mul_ps( cos0, transformC_s ), _mm_mul_ps( sin0, transformC_t ), _transformC_s, _mm_mul_ps( cos0, transformC_t ) };
			
			__m128 vecC_t = _mm_mul_ps( sinA, point5 );
			__m128 bone1B[] = { _mm_mul_ps( sin0, vecC_t ), _mm_add_ps( _mm_mul_ps( cosA, point5 ), point5 ), _mm_mul_ps( cos0, vecC_t ) };
			__m128 bone2B[] = { bone1B[ 0 ], _mm_add_ps( bone1B[ 1 ], _mm_mul_ps( transformB_t, point59 ) ), bone1B[ 2 ] };
			
			vec3 translate[] = { 
				appState->stadiumFan_position[ seat_index + 0 ],
				appState->stadiumFan_position[ seat_index + 1 ],
				appState->stadiumFan_position[ seat_index + 2 ],
				appState->stadiumFan_position[ seat_index + 3 ],
			};
			
			__m128 translate0[] = {
				_mm_set_ps( translate[ 3 ].x, translate[ 2 ].x, translate[ 1 ].x, translate[ 0 ].x ),
				_mm_set_ps( translate[ 3 ].y, translate[ 2 ].y, translate[ 1 ].y, translate[ 0 ].y ),
				_mm_set_ps( translate[ 3 ].z, translate[ 2 ].z, translate[ 1 ].z, translate[ 0 ].z ),
			};
			__m128 translate1[] = {
				_mm_add_ps( bone0[ 0 ], translate0[ 0 ] ), 
				_mm_add_ps( bone0[ 1 ], translate0[ 1 ] ), 
				_mm_add_ps( bone0[ 2 ], translate0[ 2 ] ), 
			};
			__m128 translate2[] = {
				_mm_add_ps( bone1B[ 0 ], translate0[ 0 ] ), 
				_mm_add_ps( bone1B[ 1 ], translate0[ 1 ] ), 
				_mm_add_ps( bone1B[ 2 ], translate0[ 2 ] ), 
			};
			__m128 translate3[] = {
				_mm_add_ps( bone2B[ 0 ], translate0[ 0 ] ), 
				_mm_add_ps( bone2B[ 1 ], translate0[ 1 ] ), 
				_mm_add_ps( bone2B[ 2 ], translate0[ 2 ] ), 
			};
			transformVertices(  rotation0, translate0, appState->fanMesh_position, position,  0, 16 );
			transformVertices( transformD, translate1, appState->fanMesh_position, position, 16, 32 );
			transformVertices( transformE, translate2, appState->fanMesh_position, position, 32, 48 );
			transformVertices( transformF, translate3, appState->fanMesh_position, position, 48, 72 );
			
			uint32 model_texIndex[] = {
				appState->stadiumFan_texSelect[ seat_index + 0 ],
				appState->stadiumFan_texSelect[ seat_index + 1 ],
				appState->stadiumFan_texSelect[ seat_index + 2 ],
				appState->stadiumFan_texSelect[ seat_index + 3 ],
			};
			for( int32 model_index = 0; model_index < 4; model_index++ ) {
				uint32  texIndex = model_texIndex[ model_index ];
				uint8   texIndexA = (   texIndex         & 0xFF ); // pants
				uint8   texIndexB = ( ( texIndex >>  8 ) & 0xFF ); // shirt
				uint8   texIndexC = ( ( texIndex >> 16 ) & 0xFF ); // skin
				
				int32 base_index = model_index * 72;
				for( int32 iter = 0; iter < 144; iter++ ) {
					vertex0[ vertexBuffer->temp_count ].position =                      position[ base_index + appState->fanMesh_posIndex[ iter ]                 ];
					vertex0[ vertexBuffer->temp_count ].texCoord = appState->fanMesh_texCoord[              appState->fanMesh_texIndex[ iter ] + texIndexA * 8 ];
					vertexBuffer->temp_count++;
				}
				for( int32 iter = 144; iter < 180; iter++ ) {
					vertex0[ vertexBuffer->temp_count ].position =                      position[ base_index + appState->fanMesh_posIndex[ iter ]                 ];
					vertex0[ vertexBuffer->temp_count ].texCoord = appState->fanMesh_texCoord[              appState->fanMesh_texIndex[ iter ] + texIndexB * 8 ];
					vertexBuffer->temp_count++;
				}
				for( int32 iter = 180; iter < 216; iter++ ) {
					vertex0[ vertexBuffer->temp_count ].position =                      position[ base_index + appState->fanMesh_posIndex[ iter ]                 ];
					vertex0[ vertexBuffer->temp_count ].texCoord = appState->fanMesh_texCoord[              appState->fanMesh_texIndex[ iter ] + texIndexC * 8 ];
					vertexBuffer->temp_count++;
				}
				for( int32 iter = 216; iter < 246; iter++ ) {
					vertex0[ vertexBuffer->temp_count ].position =                      position[ base_index + appState->fanMesh_posIndex[ iter ]                 ];
					vertex0[ vertexBuffer->temp_count ].texCoord = appState->fanMesh_texCoord[              appState->fanMesh_texIndex[ iter ] + texIndexC * 8 ];
					vertexBuffer->temp_count++;
				}
				for( int32 iter = 246; iter < 276; iter++ ) {
					vertex0[ vertexBuffer->temp_count ].position =                      position[ base_index + appState->fanMesh_posIndex[ iter ]                 ];
					vertex0[ vertexBuffer->temp_count ].texCoord = appState->fanMesh_texCoord[              appState->fanMesh_texIndex[ iter ] + texIndexB * 8 ];
					vertexBuffer->temp_count++;
				}
				for( int32 iter = 276; iter < 306; iter++ ) {
					vertex0[ vertexBuffer->temp_count ].position =                      position[ base_index + appState->fanMesh_posIndex[ iter ]                 ];
					vertex0[ vertexBuffer->temp_count ].texCoord = appState->fanMesh_texCoord[              appState->fanMesh_texIndex[ iter ] + texIndexC * 8 ];
					vertexBuffer->temp_count++;
				}
				for( int32 iter = 306; iter < face_count; iter++ ) {
					vertex0[ vertexBuffer->temp_count ].position =                      position[ base_index + appState->fanMesh_posIndex[ iter ]                 ];
					vertex0[ vertexBuffer->temp_count ].texCoord = appState->fanMesh_texCoord[              appState->fanMesh_texIndex[ iter ] + texIndexB * 8 ];
					vertexBuffer->temp_count++;
				}
			}
		}
	}
	renderSystem->object[ renderSystem->object_count++ ] = { mesh.drawType, vertex_index, vertex_total, Vec3( 0.0f, 0.0f, 0.0f ), scale0, Quat( Vec3( 1.0f, 0.0f, 0.0f ), 0.0f ), texture_bufferID, COLOR_WHITE };
	
	_popArray( tempMemory, vec3, position_count * 4 );
}
