
internal void
APP_updateAndRender( APP_STATE * appState, PLATFORM * platform ) {
	PROFILE_FUNCTION();
	
	RENDER_SYSTEM_2D * renderSystem2D = &platform->renderSystem2D;
	RENDER_SYSTEM_3D * renderSystem3D = &platform->renderSystem3D;
	
	MOUSE_STATE    * mouse    = &platform->mouse;
	KEYBOARD_STATE * keyboard = &platform->keyboard;
	
	MEMORY * permMemory = &platform->permMemory;
	MEMORY * tempMemory = &platform->tempMemory;
	
	ASSET_SYSTEM * assetSystem = &platform->assetSystem;
	
	flo32 court_halfLength = COURT__HALF_LENGTH;
	flo32 court_halfWidth  = COURT__HALF_WIDTH;
	
	if( !appState->isInitialized ) {
		appState->isInitialized = true;
		
		flo32  row_stride = 1.5f;
		flo32 seat_stride = 0.825f;
		
		flo32 root2 = sqrtf( 2.0f ) * 0.5f;
		flo32 edgeX = court_halfWidth  + 4.0f;
		flo32 edgeY = court_halfLength + 4.0f;
		flo32 cornerX = court_halfWidth  + root2 * 4.0f;
		flo32 cornerY = court_halfLength + root2 * 4.0f;
		flo32 offsetY = 2.0f + seat_stride * 7.0f;
		
		int32  row0_seatCount[ 10 ] = { 15, 15,  2, 18,  2, 15, 15,  2, 18,  2, };
		int32  rowN_seatCount[ 10 ] = { 15, 15, 31, 18, 31, 15, 15, 31, 18, 31, };
		
		int32 stadium_rowCount[  10 ] = { 16, 16, 12, 14, 12, 16, 16, 12, 14, 12, };
		int32 stadium_seatCount[ 10 ] = { 15, 15, -1, 18, -1, 15, 15, -1, 18, -1, };
		int32 stadium_gradientIndex[ 20 ] = { 4, 9,   4, 9,   3, 7,   3, 8,   3, 7,   4, 9,   4, 9,   3, 7,   3, 8,   3, 7, };
		
		int32 corner_seatCount[ 12 ] = {};
		for( int32 iter = 0; iter < 12; iter++ ) {
			corner_seatCount[ iter ] = ( int32 )lerp( 2.0f, ( flo32 )iter / 12.0f, 34.0f );
		}
		
		vec2 stadium_basePos[ 10 ] = {
			Vec2(  edgeX,  offsetY ), Vec2(  edgeX, -offsetY ), Vec2(  cornerX, -cornerY ), Vec2( 0.0f, -edgeY ), Vec2( -cornerX, -cornerY ), 
			Vec2( -edgeX, -offsetY ), Vec2( -edgeX,  offsetY ), Vec2( -cornerX,  cornerY ), Vec2( 0.0f,  edgeY ), Vec2(  cornerX,  cornerY ),
		};
		
		vec2 stadium_rowAxis[ 10 ] = {
			Vec2(  row_stride, 0.0f ), Vec2(  row_stride, 0.0f ), Vec2(  root2, -root2 ) * row_stride, Vec2( 0.0f, -row_stride ), Vec2( -root2, -root2 ) * row_stride,
			Vec2( -row_stride, 0.0f ), Vec2( -row_stride, 0.0f ), Vec2( -root2,  root2 ) * row_stride, Vec2( 0.0f,  row_stride ), Vec2(  root2,  root2 ) * row_stride,
		};
		vec2 stadium_seatAxis[ 10 ] = {
			Vec2( 0.0f,  seat_stride ), Vec2( 0.0f,  seat_stride ), Vec2(  root2,  root2 ) * seat_stride, Vec2(  seat_stride, 0.0f ), Vec2(  root2, -root2 ) * seat_stride,
			Vec2( 0.0f, -seat_stride ), Vec2( 0.0f, -seat_stride ), Vec2( -root2, -root2 ) * seat_stride, Vec2( -seat_stride, 0.0f ), Vec2( -root2,  root2 ) * seat_stride,
		};
		
		int32    seatPos_index = 0;
		for( int32 section_index = 0; section_index < 10; section_index++ ) {
			vec2  row_axis = stadium_rowAxis[  section_index ];
			vec2 seat_axis = stadium_seatAxis[ section_index ];
			
			vec2 basePos = stadium_basePos[ section_index ];
			
			int32 row_count = stadium_rowCount[ section_index ];
			flo32 yStride   = 2.0f / ( flo32 )row_count;
			
			flo32 row0_count = ( flo32 )( row0_seatCount[ section_index ] - 1 ) * 0.5f;
			flo32 rowN_count = ( flo32 )( rowN_seatCount[ section_index ] - 1 ) * 0.5f;
			
			vec2       fan_offset      = getNormal( -row_axis ) * 0.5f;
			vec2       row_offset      =  row_axis * 0.3333f;
			vec2      seat_offset      = seat_axis * 0.5f;
			
			int32 seat_count = stadium_seatCount[ section_index ];
			int32 * gradientIndex = stadium_gradientIndex + ( section_index * 2 );
			for( int32 row_index = 0; row_index < row_count; row_index++ ) {
				if( ( section_index == 2 ) || ( section_index == 4 ) || ( section_index == 7 ) || ( section_index == 9 ) ) {
					seat_count = corner_seatCount[ row_index ];
				}
				
				flo32 yPos = yStride * ( flo32 )row_index;
				if( row_index > gradientIndex[ 0 ] ) {
					yPos += ( yStride * ( flo32 )( row_index - gradientIndex[ 0 ] ) );
				}
				if( row_index > gradientIndex[ 1 ] ) {
					yPos += ( yStride * ( flo32 )( row_index - gradientIndex[ 1 ] ) );
				}
				
				vec2 atPos = basePos  + row_axis * ( flo32 )row_index - seat_axis * ( ( flo32 )( seat_count - 1 ) * 0.5f );
				
				vec2  fanPos = atPos + fan_offset;
				for( int32 seat_index = 0; seat_index < seat_count; seat_index++ ) {
					appState->stadiumFan_position[ seatPos_index++ ] = Vec3( fanPos.x, yPos, fanPos.y );
					fanPos += seat_axis;
				}
			}
			
			flo32 radians[ 8 ] = { -( PI * 0.5f ), -( PI * 0.25f ), 0.0f, ( PI * 0.25f ), ( PI * 0.5f ), ( PI * 0.75f ), PI, ( PI * 1.25f ), };
			for( int32 iter = 0; iter < 8; iter++ ) {
				appState->stadiumFan_rotation[ iter ] = radians[ iter ];
			}
		}
		
		genStadiumFan( appState );
		#if DEBUG_BUILD
		// genStadiumSeat();
		// genStadiumSeats( court_halfWidth, court_halfLength );
		// genJumbotron();
		// genHoopCollision();
		// genPlayerWalkBound( stadium_rowAxis, stadium_seatAxis, stadium_basePos, stadium_rowCount, row0_seatCount, rowN_seatCount );
		// genStadium( stadium_rowAxis, stadium_seatAxis, stadium_basePos, stadium_rowCount, row0_seatCount, rowN_seatCount, stadium_seatCount, stadium_gradientIndex, corner_seatCount );
		#endif // DEBUG_BUILD
		
		appState->currentFan_boundPos = Vec3( 1000.0f, 0.0f, 1000.0f );
		appState->currentFan_boundXAxis = Vec3( 1.0f, 0.0f, 0.0f );
		appState->currentFan_boundZAxis = Vec3( 0.0f, 0.0f, 1.0f );
		
		appState->mesh_stadiumSeat  = addPermMesh( renderSystem3D, assetID_mesh_stadiumSeat  );
		appState->mesh_stadiumSeats = addPermMesh( renderSystem3D, assetID_mesh_stadiumSeats );
		
		appState->bone[ 0 ] = Vec3( 0.0f, 0.5f,  0.0f );
		appState->bone[ 1 ] = Vec3( 0.0f, 1.0f,  0.0f );
		appState->bone[ 2 ] = Vec3( 0.0f, 1.59f, 0.0f );
		
		for( int32 iter = 0; iter < 2248; iter++ ) {
			// appState->stadiumFan_timer[     iter ] = ( flo32 )( rand() % 1000 ) * 0.001f;
			appState->stadiumFan_texSelect[ iter ] = ( ( uint8 )( rand() % 8 ) << 16 ) | ( ( uint8 )( rand() % 8 ) << 8 ) | ( uint8 )( rand() % 8 );
			appState->stadiumFan_radians[   iter ] = PI * 2.0f;
			appState->stadiumFan_t[         iter ] = 1.0f;
		}
		
		appState->player_isHoldingBall = true;
		appState->player_shotPower     = 0.25f;
		
		appState->jumbotron_timer = 300.0f;
		
		#define COLUMN_COUNT  ( 15 )
		int32 column_count = COLUMN_COUNT;
		int32 seatInColumn[       COLUMN_COUNT ] = {};
		int32 seatInColumn_base[  COLUMN_COUNT ] = {};
		int32 seat_count[ 12 ] = { 2, 4, 7, 10, 12, 15, 18, 20, 23, 26, 28, 31, };
		int32 corner_column[ 196 ] = {};
		int32 corner_columnIndex = 0;
		
		for( int32 row_index = 0; row_index < 12; row_index++ ) {
			flo32 stride = ( flo32 )column_count / ( flo32 )seat_count[ row_index ];
			flo32 at = 0.5f;
			for( int32 seat_index = 0; seat_index < seat_count[ row_index ]; seat_index++ ) {
				int32 index = ( int32 )at;
				index = minValue( index, 14 );
				seatInColumn[ index ]++;
				corner_column[ corner_columnIndex + ( seat_count[ row_index ] - 1 ) - seat_index ] = index;
				at += stride;
			}
			corner_columnIndex += seat_count[ row_index ];
		}
		
		int32 atIndex = 0;
		for( int32 iter = 0; iter < COLUMN_COUNT; iter++ ) {
			seatInColumn_base[ iter ] = atIndex;
			atIndex += seatInColumn[ iter ];
		}
		
		{ // reorganize stadiumFan_positions by column
			int32 fanWave_index = 0;
			int32 fanWave_column_count = 0;
			
			vec3 stadiumFan_position[ 2248 ] = {};
			memcpy( stadiumFan_position, appState->stadiumFan_position, sizeof( vec3 ) * 2248 );
			
			int32 fan_count[] = { 240, 240, 196, 252, 196, 240, 240, 196, 252, 196 };
			int32 counter = 0;
		
			int32 nRow[ 10 ] = { 16, 16, 12, 14, 12, 16, 16, 12, 14, 12, };
			int32 nCol[ 10 ] = { 15, 15, -1, 18, -1, 15, 15, -1, 18, -1, };
			
			int32 min_index[ 10 ] = {};
			int32 max_index[ 10 ] = {};
			for( int32 iter = 0; iter < 10; iter++ ) {
				min_index[ iter ] = counter;
				counter += fan_count[ iter ];
				max_index[ iter ] = counter;
			}
			
			int32 sum = 0;
			for( int32 iter = 0; iter < column_count; iter++ ) {
				sum += seatInColumn[ iter ];
			}
			
			for( int32 section_index = 0; section_index < 10; section_index++ ) {
				int32 min0 = min_index[ section_index ];
				int32 max0 = max_index[ section_index ];
				if( ( section_index == 2 ) || ( section_index == 4 ) || ( section_index == 7 ) || ( section_index == 9 ) ) {
					int32 seatInColumn_count[ COLUMN_COUNT ] = {};
					for( int32 seat_index = min0; seat_index < max0; seat_index++ ) {
						int32 colIndex = corner_column[ seat_index - min0 ];
						int32 destIndex = seatInColumn_base[ colIndex ] + seatInColumn_count[ colIndex ];
						Assert( destIndex < ( max0 - min0 ) );
						appState->stadiumFan_position[ min0 + destIndex ] = stadiumFan_position[ seat_index ];
						seatInColumn_count[ colIndex ]++;
					}
					for( int32 column_index = 0; column_index < 15; column_index++ ) {
						appState->fanWave_column_baseIndex[  fanWave_index ] = fanWave_column_count;
						appState->fanWave_column_count[ fanWave_index ] = seatInColumn[ column_index ];
						fanWave_index++;
						fanWave_column_count += seatInColumn[ column_index ];
					}
				} else {
					int32 nRow0 = nRow[ section_index ];
					int32 nCol0 = nCol[ section_index ];
					for( int32 seat_index = min0; seat_index < max0; seat_index++ ) {
						int32 row = ( seat_index - min0 ) / nCol0;
						int32 col = ( seat_index - min0 ) % nCol0;
						      col = nCol0 - col - 1;
								
						appState->stadiumFan_position[ min0 + col * nRow0 + row ] = stadiumFan_position[ seat_index ];
					}
					for( int32 column_index = 0; column_index < nCol0; column_index++ ) {
						appState->fanWave_column_baseIndex[  fanWave_index ] = fanWave_column_count;
						appState->fanWave_column_count[ fanWave_index ] = nRow0;
						fanWave_index++;
						fanWave_column_count += nRow0;
					}
				}
			}
			
			uint32 breakHere = 10;
		}
		
		for( int32 fan_index = 0; fan_index < 2248; fan_index++ ) {
			appState->stadiumFan_rate[ fan_index ] = 1.0f;
			appState->fanCheer_buffer[ fan_index ] = fan_index;
		}
		
		appState->camera_yAxis = Vec3( 0.0f, 1.0f, 0.0f );
		appState->camera_zAxis = Vec3( 0.0f, 0.0f, 1.0f );
		
		appState->ball_contactPos[ 0 ] = Vec3( 0.0f, -1.0f, 0.0f );
		appState->ball_contactPos[ 1 ] = Vec3( 0.0f, -1.0f, 0.0f );
		appState->ball_contactNormal[ 0 ] = Vec3( 0.0f, 1.0f, 0.0f );
		appState->ball_contactNormal[ 1 ] = Vec3( 0.0f, 1.0f, 0.0f );
		
		appState->scoreState_home.position = Vec3( 0.0f, 0.0f, -HOOP__DIST_FROM_CENTER ) + Vec3( 0.0f, 3.2f,  0.275f );
		appState->scoreState_away.position = Vec3( 0.0f, 0.0f,  HOOP__DIST_FROM_CENTER ) + Vec3( 0.0f, 3.2f, -0.275f );
		
		appState->helpTutorial_counter = 10;
		
		initFanWaveTimer( appState );
	}
	
	appState->jumbotron_timer -= platform->targetSec;
	
	flo32 player_height = 1.524f;
	flo32 ball_radius = 0.2413f * 0.5f;
	vec2 player_position = appState->player_position;
	flo32 player_yPos = getYPos( assetSystem, appState->player_position );
	vec3 camera_position = Vec3( player_position.x, player_yPos + player_height, player_position.y );
	if( wasPressed( mouse, mouseButton_left ) ) {
		if( appState->player_isHoldingBall ) {
			appState->player_isHoldingBall = false;
			appState->ball_velocity = getNormal( ( appState->camera_yAxis * 0.5f ) - appState->camera_zAxis ) * ( appState->player_shotPower * 17.575f + 2.0f );
			appState->ball_rotateVelocity = ( flo32 )( rand() % 101 ) / 100.0f * PI;
		}
	}
	if( wasPressed( mouse, mouseButton_right ) ) {
		appState->player_isHoldingBall = true;
	}
	
#if DEBUG_BUILD
	if( wasPressed( keyboard, keyCode_t ) ) {
		initTestShot( appState, player_height, ball_radius );
	}
	
	flo32 vRadians = appState->player_shotPower * ( PI * 0.5f );
	flo32 vDegrees = vRadians / ( 2.0f * PI ) * 360.0f;
	vec3  vVector  = Vec3( 0.0f, sinf( vRadians ), -cosf( vRadians ) );
	debug_line( appState->shot_position, appState->shot_position + vVector, COLOR_CYAN );
	
	debug_line( appState->shot_position, appState->shot_position + appState->shot_xAxis,    COLOR_RED     );
	debug_line( appState->shot_position, appState->shot_position + appState->shot_yAxis,    COLOR_GREEN   );
	debug_line( appState->shot_position, appState->shot_position + appState->shot_zAxis,    COLOR_BLUE    );
	debug_line( appState->shot_ballPos,  appState->shot_ballPos  + appState->shot_velocity, COLOR_MAGENTA );
#endif // DEBUG_BUILD
	
	updatePlayerMovement( appState, assetSystem, platform->targetSec, 0.35f );
	
	flo32 player_hAngle = appState->player_hAngle;
	flo32 player_vAngle = appState->player_vAngle;
	vec2 camera_moveVector = ( appState->cursor_position - appState->app_halfDim ) * 0.0015f;
	player_hAngle -= camera_moveVector.x;
	player_vAngle += camera_moveVector.y;
	
	player_vAngle = maxValue( player_vAngle, -( PI * 0.5f ) );
	player_vAngle = minValue( player_vAngle,  ( PI * 0.5f ) );
	
	quat court_orientation = Quat( Vec3( 1.0f, 0.0f, 0.0f ), -( PI * 0.5f ) ) * Quat( Vec3( 0.0f, 0.0f, 1.0f ), PI * 0.5f );
	vec3 court_scale       = Vec3( COURT__HALF_LENGTH, COURT__HALF_WIDTH, 1.0f );
	
	vec2  vector = Vec2( -sinf( player_hAngle ), -cosf( player_hAngle ) ) * sinf( player_vAngle );
	vec3 camera_xAxis     = Vec3( cosf( player_hAngle ), 0.0f, -sinf( player_hAngle ) );
	vec3 camera_yAxis     = getNormal( Vec3( vector.x, cosf( player_vAngle ), vector.y ) );
	vec3 camera_zAxis     = getNormal( cross( camera_xAxis, camera_yAxis ) );
	renderSystem3D->camera_transform = genCameraTransform( camera_position, camera_xAxis, camera_yAxis, camera_zAxis );
	
	vec2 player_hAxis = getNormal( Vec2( camera_xAxis.x,  camera_xAxis.z ) );
	vec2 player_vAxis = getNormal( Vec2( player_hAxis.y, -player_hAxis.x ) );
	
	appState->player_hAngle = player_hAngle;
	appState->player_vAngle = player_vAngle;
	appState->player_hAxis = player_hAxis;
	appState->player_vAxis = player_vAxis;
	appState->camera_yAxis = camera_yAxis;
	appState->camera_zAxis = camera_zAxis;
	
	if( appState->player_isHoldingBall ) {
		appState->ball_position    = camera_position - getNormal( camera_zAxis + camera_yAxis ) * ( 0.5f * 0.3048f + ball_radius );
		
		appState->ball_rotateVelocity = 0.0f;
		appState->ball_radians        = 0.0f;
		appState->ball_rotateAxis  = getNormal( camera_xAxis );
		
		appState->ball_orientation = Quat( Vec3( 0.0f, 1.0f, 0.0f ), player_hAngle ) * Quat( Vec3( 1.0f, 0.0f, 0.0f ), -( player_vAngle + PI / 4.0f ) );
	} else {
		updateBallCollision( appState, assetSystem, platform->targetSec );
	}
	vec3  ball_scale    = Vec3( ball_radius, ball_radius, ball_radius );
	quat ball_orientation = Quat( appState->ball_rotateAxis, appState->ball_radians ) * appState->ball_orientation;
	debug_line( appState->ball_position, appState->ball_position + appState->ball_rotateAxis, COLOR_MAGENTA );
	drawMesh( renderSystem3D, assetID_mesh_centerPanel,    assetID_texture_basketballCourt, Vec3( 0.0f, 0.0f, 0.0f ),   court_scale, court_orientation, COLOR_WHITE );
	drawMesh( renderSystem3D, assetID_mesh_sphere,         assetID_texture_basketball,      appState->ball_position,    ball_scale, ball_orientation, COLOR_WHITE );
	
	flo32 hoop_distFromCenter = 13.1064f;
	drawMesh( renderSystem3D, assetID_mesh_basketballHoop, assetID_texture_basketballHoop, Vec3( 0.0f, 0.0f, -hoop_distFromCenter ), Vec3( 1.0f, 1.0f, 1.0f ), Quat( Vec3( 1.0f, 0.0f, 0.0f ), -( PI * 0.5f ) ),                                        COLOR_WHITE );
	drawMesh( renderSystem3D, assetID_mesh_basketballHoop, assetID_texture_basketballHoop, Vec3( 0.0f, 0.0f,  hoop_distFromCenter ), Vec3( 1.0f, 1.0f, 1.0f ), Quat( Vec3( 1.0f, 0.0f, 0.0f ), -( PI * 0.5f ) ) * Quat( Vec3( 0.0f, 0.0f, 1.0f ), PI ), COLOR_WHITE );
	
	updateScore( appState, &appState->scoreState_home, appState->ball_position );
	updateScore( appState, &appState->scoreState_away, appState->ball_position );
	
	drawMesh( renderSystem3D, assetID_mesh_stadiumFloor, assetID_texture_stadiumFloor, Vec3( 0.0f, 0.0f, 0.0f ), Vec3( 1.0f, 1.0f, 1.0f ), QuatDef(), COLOR_WHITE );
	
	flo32 radiansPerSecond = PI * 1.2f * platform->targetSec;
	flo32 maxDist = 3.0f;
	flo32 minDist = 0.5f;
	for( int32 seat_index = 0; seat_index < 2248; seat_index++ ) {
		vec3 vector0 = appState->ball_position - appState->stadiumFan_position[ seat_index ];
		vec3 normal0 = getNormal( vector0 );
		vec3 vector1 = Vec3( 0.0f, 1.0f, 0.0f );
		
		flo32 dot01 = maxValue( dot( normal0, vector1 ), 0.0f );
		flo32 dist0 = lerp( minDist, dot01, maxDist );
		flo32 distSq0 = dist0 * dist0;
		
		flo32 distSq = getLengthSq( vector0 );
		if( ( distSq < distSq0 ) && ( appState->stadiumFan_radians[ seat_index ] > ( PI * 2.0f ) ) ) {
			appState->stadiumFan_radians[ seat_index ] = 0.0f;
			appState->stadiumFan_rate[ seat_index ] = 1.2f;
		}
	}
	animateFans( renderSystem3D, appState, tempMemory );
	
	drawPermMesh( renderSystem3D, appState->mesh_stadiumSeats, assetID_texture_stadiumSeat, Vec3( 0.0f, 0.0f, 0.0f ), Vec3( 1.0f, 1.0f, 1.0f ), QuatDef(), COLOR_WHITE );
	
	flo32 root2 = sqrtf( 2.0f ) * 0.5f;
	flo32  row_stride = 1.5f;
	flo32 seat_stride = 0.825f;
	flo32 heightA = 1.4f;
	flo32 heightB = 1.9f;
	flo32 xLengthSq = seat_stride * seat_stride;
	flo32 zLengthSq =  row_stride *  row_stride;
	
	boo32 currentFan_inBound = false;
	{
		vec3  currentFan_pos = appState->currentFan_boundPos;
		flo32 currentFan_t   = appState->stadiumFan_t[ appState->currentFan_index ];
		
		flo32 yLength   = lerp( heightB, appState->stadiumFan_t[ appState->currentFan_index ], heightA );
		flo32 yLengthSq = yLength * yLength;
		
		vec3 xAxis0 = appState->currentFan_boundXAxis;
		vec3 yAxis0 = Vec3( 0.0f, yLength, 0.0f );
		vec3 zAxis0 = appState->currentFan_boundZAxis;
		
		vec3 vector = appState->ball_position - currentFan_pos;
		
		flo32 xDot = dot( vector, xAxis0 ) / xLengthSq;
		flo32 yDot = dot( vector, yAxis0 ) / yLengthSq;
		flo32 zDot = dot( vector, zAxis0 ) / zLengthSq;
		
		currentFan_inBound = ( xDot >= 0.0f ) && ( xDot <= 1.0f ) && ( yDot >= 0.0f ) && ( yDot <= 1.0f ) && ( zDot >= 0.0f ) && ( zDot <= 1.0f );
	}
	if( !currentFan_inBound ) {
		int32 fan_count[] = { 240, 240, 196, 252, 196, 240, 240, 196, 252, 196 };
		int32 counter = 0;
		
		int32 minIndex[ 10 ] = {};
		int32 maxIndex[ 10 ] = {};
		for( int32 iter = 0; iter < 10; iter++ ) {
			minIndex[ iter ] = counter;
			counter += fan_count[ iter ];
			maxIndex[ iter ] = counter;
		}
		
		vec2 xAxis[ 10 ] = {
			Vec2( 0.0f,  seat_stride ), Vec2( 0.0f,  seat_stride ), Vec2(  root2,  root2 ) * seat_stride, Vec2(  seat_stride, 0.0f ), Vec2(  root2, -root2 ) * seat_stride,
			Vec2( 0.0f, -seat_stride ), Vec2( 0.0f, -seat_stride ), Vec2( -root2, -root2 ) * seat_stride, Vec2( -seat_stride, 0.0f ), Vec2( -root2,  root2 ) * seat_stride,
		};
		vec2 zAxis[ 10 ] = {
			Vec2(  row_stride, 0.0f ), Vec2(  row_stride, 0.0f ), Vec2(  root2, -root2 ) * row_stride, Vec2( 0.0f, -row_stride ), Vec2( -root2, -root2 ) * row_stride,
			Vec2( -row_stride, 0.0f ), Vec2( -row_stride, 0.0f ), Vec2( -root2,  root2 ) * row_stride, Vec2( 0.0f,  row_stride ), Vec2(  root2,  root2 ) * row_stride,
		};
		for( int32 section_index = 0; section_index < 10; section_index++ ) {
			vec3 xAxis0 = Vec3( xAxis[ section_index ].x, 0.0f, xAxis[ section_index ].y );
			vec3 zAxis0 = Vec3( zAxis[ section_index ].x, 0.0f, zAxis[ section_index ].y );
			vec3 seat_offset = -xAxis0 * 0.5f - zAxis0 * 0.3333f;
			
			for( int32 seat_index = minIndex[ section_index ]; seat_index < maxIndex[ section_index ]; seat_index++ ) {
				vec3 seat_pos = appState->stadiumFan_position[ seat_index ] + seat_offset;
				
				flo32 yLength   = lerp( heightB, appState->stadiumFan_t[ seat_index ], heightA );
				flo32 yLengthSq = yLength * yLength;
				vec3  yAxis0 = Vec3( 0.0f, yLength, 0.0f );
				
				vec3 vector = appState->ball_position - seat_pos;
				
				flo32 xDot = dot( vector, xAxis0 ) / xLengthSq;
				flo32 yDot = dot( vector, yAxis0 ) / yLengthSq;
				flo32 zDot = dot( vector, zAxis0 ) / zLengthSq;
				
				boo32 inBound = ( xDot >= 0.0f ) && ( xDot <= 1.0f ) && ( yDot >= 0.0f ) && ( yDot <= 1.0f ) && ( zDot >= 0.0f ) && ( zDot <= 1.0f );
				if( inBound ) {
					appState->currentFan_boundPos = seat_pos;
					appState->currentFan_boundXAxis = xAxis0;
					appState->currentFan_boundZAxis = zAxis0;
					
					flo32 hRadians = ( flo32 )( rand() % 1000 ) / 1000.0f * 2.0f * PI;
					flo32 vRadians = ( flo32 )( rand() % 1000 ) / 1000.0f * PI * 0.3333f + PI * 0.1667f;
					flo32 velocity = 8.0f + ( flo32 )( rand() % 1000 ) / 1000.0f * 8.0f;
					
					appState->currentFan_velocity = Vec3( cosf( hRadians ) * cosf( vRadians ), sinf( vRadians ), sinf( hRadians ) * cosf( vRadians ) ) * velocity;
					appState->ball_velocity = appState->currentFan_velocity;
				}
			}
		}
	} else {
		appState->ball_velocity = appState->currentFan_velocity;
	}
	
	drawMesh( renderSystem3D, assetID_mesh_jumbotron, assetID_texture_jumbotron );
	drawJumbotronText( renderSystem3D, appState );
	
	flo32  scale_ratio[ 2 ] = { 1.0f, 0.75f };
	ASSET_ID textureID[ 2 ] = { assetID_texture_markerA, assetID_texture_markerB };
	
	for( int32 iter = 0; iter < 2; iter++ ) {
		vec3 normal = appState->ball_contactNormal[ iter ];
		flo32 phi   =  asinf( normal.y );
		flo32 theta = atan2f( normal.x, normal.z );
		
		flo32 dist = getLength( appState->ball_contactPos[ iter ] - camera_position );
		
		flo32 scale = lerp( 0.025f, dist / 40.0f, 0.4f ) * scale_ratio[ iter ];
		quat orientationA = Quat( Vec3( 0.0f, 1.0f, 0.0f ), theta      ) * Quat( Vec3( 1.0f, 0.0f, 0.0f ), -phi );
		quat orientationB = Quat( Vec3( 0.0f, 1.0f, 0.0f ), theta + PI ) * Quat( Vec3( 1.0f, 0.0f, 0.0f ),  phi );
		
		drawMesh( renderSystem3D, assetID_mesh_circlePanel, textureID[ iter ], appState->ball_contactPos[ iter ], Vec3( scale, scale, scale ), orientationA, COLOR_WHITE );
		drawMesh( renderSystem3D, assetID_mesh_circlePanel, textureID[ iter ], appState->ball_contactPos[ iter ], Vec3( scale, scale, scale ), orientationB, COLOR_WHITE );
	}
	
	updateFanWaveTimer( appState, platform->targetSec );
	updateFanWave( appState, platform->targetSec );
	updateFanCheer( appState, radiansPerSecond );
	
	rect app_bound   = appState->app_bound;
	vec2 app_halfDim = appState->app_halfDim;
	
	vec2 shotPower_dim   = Vec2( 420.0f, 90.0f );
	rect shotPower_bound = rectBCD( Vec2( app_halfDim.x, 10.0f ), shotPower_dim );
	drawRect( renderSystem2D, assetID_texture_shotPower_bar, shotPower_bound, COLOR_WHITE );
	
	vec2 bar_offset = Vec2( 16.0f, 14.0f );
	vec2 bar_dim    = Vec2( getWidth( shotPower_bound ) - ( bar_offset.x * 2.0f ), 42.0f );
	rect bar_bound  = rectBLD( getBL( shotPower_bound ) + Vec2( bar_offset.x, bar_offset.y ), bar_dim );
	
	int32 notch_count = ( int32 )( appState->player_shotPower * 100.0f );
	
	vec2 atPos     = getBL( bar_bound );
	vec2 notch_dim = Vec2( bar_dim.x / 100.0f, bar_dim.y );
	for( int32 notch_index = 0; notch_index < notch_count; notch_index++ ) {
		rect bound = rectBLD( atPos, notch_dim );
		drawRect( renderSystem2D, assetID_texture_shotPower_notch, bound, COLOR_WHITE );
		atPos.x += notch_dim.x;
	}
	
	vec2 helpButton_dim = Vec2( 54.0f, 54.0f );
	rect helpButton_bound = rectBRD( getBL( shotPower_bound ) + Vec2( 4.0f, 8.0f ), helpButton_dim );
	drawRect( renderSystem2D, assetID_texture_controlsHelp_button, helpButton_bound, COLOR_WHITE );
	
	if( appState->helpTutorial_counter > 0 ) {
		flo32 targetTime = 1.0f;
		
		flo32 t = clamp01( 1.0f - ( appState->helpTutorial_timer / targetTime ) );
		
		vec2 help_dim   = Vec2( 259.0f, 37.0f );
		rect help_bound = rectCD( Vec2( app_halfDim.x, app_halfDim.y * 1.3333f ), help_dim );
		drawRect( renderSystem2D, assetID_texture_controlsHelp_text, help_bound, COLOR_WHITE * t );
		
		appState->helpTutorial_timer += platform->targetSec;
		if( appState->helpTutorial_timer >= targetTime ) {
			appState->helpTutorial_counter--;
			appState->helpTutorial_timer = 0.0f;
		}
	}
	if( appState->showControls ) {
		drawRect( renderSystem2D, assetID_texture_whiteTexture, app_bound, Vec4( 0.65f, 0.65f, 0.65f, 0.65f ) );
		
		vec2 exit_dim   = Vec2( 158.0f, 66.0f );
		rect exit_bound = rectBRD( app_halfDim + Vec2( -160.0f, 110.0f ), exit_dim );
		drawRect( renderSystem2D, assetID_texture_controlsHelp_exit, exit_bound, COLOR_WHITE );
		
		vec2 heading_dim   = Vec2( 147.0f, 39.0f ) * 1.25f;
		rect heading_bound = rectTLD( getTL( app_bound ) + Vec2( 140.0f, -60.0f ), heading_dim );
		drawRect( renderSystem2D, assetID_texture_controlsHelp_heading, heading_bound, COLOR_WHITE );
		
		vec2 controls_dim = Vec2( 833.0f, 243.0f ) * 0.9f;
		rect controls_bound = rectCD( app_halfDim + Vec2( 0.0f, -40.0f ), controls_dim );
		drawRect( renderSystem2D, assetID_texture_controlsHelp_controls, controls_bound, COLOR_WHITE );
	}
}