
#ifdef	STD_INCLUDE_DECL

#define COLLISION__COLLISION_EPSILON  ( 0.0001f )
#define COLLISION__VELOCITY_EPSILON   ( 0.00000001f )
#define COLLISION__TMIN_EPSILON       ( 0.01f )
#define COLLISION__DIST_SQ_EPSILON    ( 0.000001f )
#define COLLISION__VALID_EPSILON      ( 0.01f )

struct COLLISION_3D {
	boo32 isValid;
	flo32 t;
	vec3  normal;
};

struct COLLISION_2D {
	boo32 isValid;
	flo32 t;
	vec2  normal;
};

struct cylinder {
	vec3  origin;
	vec3  vector;
	flo32 radius;
};

struct sphere {
	vec3  origin;
	flo32 radius;
};

struct torus {
	vec3 origin;
	vec3 normal;
	flo32 major_radius;
	flo32 minor_radius;
};

struct COLLISION_GEOMETRY {
	int32  quad_count;
	int32   tri_count;
	int32  edge_count;
	int32 point_count;
	
	int32 cylinder_count;
	int32    torus_count;
};

#define COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE  ( 2048 * 4 * 3 )
struct COLLISION_GEOMETRY__TEMP_BUFFER {
	COLLISION_GEOMETRY geometry;
	flo32     quad_data[ COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE ];
	flo32      tri_data[ COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE ];
	flo32     edge_data[ COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE ];
	flo32    point_data[ COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE ];
	flo32 cylinder_data[ COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE ];
	flo32    torus_data[ COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE ];
};

#endif	// STD_INCLUDE_DECL
#ifdef	STD_INCLUDE_FUNC

internal cylinder
Cylinder( vec3 origin, vec3 vector, flo32 radius ) {
	cylinder result = { origin, vector, radius };
	return result;
}

internal torus
Torus( vec3 origin, vec3 normal, flo32 majorRadius, flo32 minorRadius ) {
	torus result = { origin, normal, majorRadius, minorRadius };
	return result;
}

//----------
// collision geometry functions
//----------

internal void
addQuad( COLLISION_GEOMETRY * collisionGeometry, MEMORY * memory, vec3 pointA, vec3 pointB, vec3 pointC, vec3 pointD ) {
	collisionGeometry->quad_count++;
	_writem( memory, vec3, pointA );
	_writem( memory, vec3, pointB );
	_writem( memory, vec3, pointC );
	_writem( memory, vec3, pointD );
}

internal void
addTri( COLLISION_GEOMETRY * collisionGeometry, MEMORY * memory, vec3 pointA, vec3 pointB, vec3 pointC ) {
	collisionGeometry->tri_count++;
	_writem( memory, vec3, pointA );
	_writem( memory, vec3, pointB );
	_writem( memory, vec3, pointC );
}

internal void
addEdge( COLLISION_GEOMETRY * collisionGeometry, MEMORY * memory, vec3 pointA, vec3 pointB ) {
	collisionGeometry->edge_count++;
	_writem( memory, vec3, pointA );
	_writem( memory, vec3, pointB );
}

internal void
addPoint( COLLISION_GEOMETRY * collisionGeometry, MEMORY * memory, vec3 pointA ) {
	collisionGeometry->point_count++;
	_writem( memory, vec3, pointA );
}

internal void
addCylinder( COLLISION_GEOMETRY * collisionGeometry, MEMORY * memory, vec3 origin, vec3 vector, flo32 radius ) {
	collisionGeometry->cylinder_count++;
	_writem( memory, vec3,  origin );
	_writem( memory, vec3,  vector );
	_writem( memory, flo32, radius );
}

internal void
addTorus( COLLISION_GEOMETRY * collisionGeometry, MEMORY * memory, vec3 origin, vec3 vector, flo32 major_radius, flo32 minor_radius ) {
	collisionGeometry->torus_count++;
	_writem( memory, vec3,  origin );
	_writem( memory, vec3,  vector );
	_writem( memory, flo32, major_radius );
	_writem( memory, flo32, minor_radius );
}

//----------
// collision geometry temp buffer functions
//----------

internal int32
collisionGeometry_getDataSize( int32 quad_count, int32 tri_count, int32 edge_count, int32 point_count, int32 cylinder_count, int32 torus_count ) {
	int32 result = 0;
	result += ( sizeof( flo32 ) * 12 *     quad_count );
	result += ( sizeof( flo32 ) *  9 *      tri_count );
	result += ( sizeof( flo32 ) *  6 *     edge_count );
	result += ( sizeof( flo32 ) *  3 *    point_count );
	result += ( sizeof( flo32 ) *  7 * cylinder_count );
	result += ( sizeof( flo32 ) *  8 *    torus_count );
	return result;
}

internal int32
getSize3D( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer ) {
	int32 result = sizeof( COLLISION_GEOMETRY );
	result += ( sizeof( flo32 ) * 12 * tempBuffer->geometry.quad_count     );
	result += ( sizeof( flo32 ) *  9 * tempBuffer->geometry.tri_count      );
	result += ( sizeof( flo32 ) *  6 * tempBuffer->geometry.edge_count     );
	result += ( sizeof( flo32 ) *  3 * tempBuffer->geometry.point_count    );
	result += ( sizeof( flo32 ) *  7 * tempBuffer->geometry.cylinder_count );
	result += ( sizeof( flo32 ) *  8 * tempBuffer->geometry.torus_count    );
	return result;
}

internal int32
getSize2D( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer ) {
	int32 result = sizeof( COLLISION_GEOMETRY );
	result += ( sizeof( flo32 ) * 8 * tempBuffer->geometry.quad_count     );
	result += ( sizeof( flo32 ) * 6 * tempBuffer->geometry.tri_count      );
	result += ( sizeof( flo32 ) * 4 * tempBuffer->geometry.edge_count     );
	result += ( sizeof( flo32 ) * 2 * tempBuffer->geometry.point_count    );
	result += ( sizeof( flo32 ) * 5 * tempBuffer->geometry.cylinder_count );
	result += ( sizeof( flo32 ) * 6 * tempBuffer->geometry.torus_count    );
	return result;
}

internal void
packGeometry3D( MEMORY * memory, COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer ) {
	COLLISION_GEOMETRY temp = tempBuffer->geometry;
	_writem( memory, COLLISION_GEOMETRY, temp );
	_writeb( memory, tempBuffer->quad_data,     sizeof( flo32 ) * 12 * temp.quad_count     );
	_writeb( memory, tempBuffer->tri_data,      sizeof( flo32 ) *  9 * temp.tri_count      );
	_writeb( memory, tempBuffer->edge_data,     sizeof( flo32 ) *  6 * temp.edge_count     );
	_writeb( memory, tempBuffer->point_data,    sizeof( flo32 ) *  3 * temp.point_count    );
	_writeb( memory, tempBuffer->cylinder_data, sizeof( flo32 ) *  7 * temp.cylinder_count );
	_writeb( memory, tempBuffer->torus_data,    sizeof( flo32 ) *  8 * temp.torus_count    );
}

internal void
packGeometry2D( MEMORY * memory, COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer ) {
	COLLISION_GEOMETRY temp = tempBuffer->geometry;
	_writem( memory, COLLISION_GEOMETRY, temp );
	_writeb( memory, tempBuffer->quad_data,     sizeof( flo32 ) * 8 * temp.quad_count     );
	_writeb( memory, tempBuffer->tri_data,      sizeof( flo32 ) * 6 * temp.tri_count      );
	_writeb( memory, tempBuffer->edge_data,     sizeof( flo32 ) * 4 * temp.edge_count     );
	_writeb( memory, tempBuffer->point_data,    sizeof( flo32 ) * 2 * temp.point_count    );
	_writeb( memory, tempBuffer->cylinder_data, sizeof( flo32 ) * 5 * temp.cylinder_count );
	_writeb( memory, tempBuffer->torus_data,    sizeof( flo32 ) * 6 * temp.torus_count    );
}

internal void
addQuad( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer, vec3 pointA, vec3 pointB, vec3 pointC, vec3 pointD ) {
	Assert( ( ( tempBuffer->geometry.quad_count + 1 ) * 12 ) < COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE );
	uint8 * ptr = ( uint8 * )( tempBuffer->quad_data + ( tempBuffer->geometry.quad_count * 12 ) );
	_write( ptr, vec3, pointA );
	_write( ptr, vec3, pointB );
	_write( ptr, vec3, pointC );
	_write( ptr, vec3, pointD );
	tempBuffer->geometry.quad_count++;
}

internal void
addTri( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer, vec3 pointA, vec3 pointB, vec3 pointC ) {
	Assert( ( ( tempBuffer->geometry.tri_count + 1 ) * 9 ) < COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE );
	uint8 * ptr = ( uint8 * )( tempBuffer->tri_data + ( tempBuffer->geometry.tri_count * 9 ) );
	_write( ptr, vec3, pointA );
	_write( ptr, vec3, pointB );
	_write( ptr, vec3, pointC );
	tempBuffer->geometry.tri_count++;
}

internal void
addEdge( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer, vec3 pointA, vec3 pointB ) {
	Assert( ( ( tempBuffer->geometry.edge_count + 1 ) * 6 ) < COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE );
	uint8 * ptr = ( uint8 * )( tempBuffer->edge_data + ( tempBuffer->geometry.edge_count * 6 ) );
	_write( ptr, vec3, pointA );
	_write( ptr, vec3, pointB );
	tempBuffer->geometry.edge_count++;
}

internal void
addEdge( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer, vec2 pointA, vec2 pointB ) {
	Assert( ( ( tempBuffer->geometry.edge_count + 1 ) * 4 ) < COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE );
	uint8 * ptr = ( uint8 * )( tempBuffer->edge_data + ( tempBuffer->geometry.edge_count * 4 ) );
	_write( ptr, vec2, pointA );
	_write( ptr, vec2, pointB );
	tempBuffer->geometry.edge_count++;
}

internal void
addPoint( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer, vec3 pointA ) {
	Assert( ( ( tempBuffer->geometry.point_count + 1 ) * 3 ) < COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE );
	uint8 * ptr = ( uint8 * )( tempBuffer->point_data + ( tempBuffer->geometry.point_count * 3 ) );
	_write( ptr, vec3, pointA );
	tempBuffer->geometry.point_count++;
}

internal void
addPoint( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer, vec2 pointA ) {
	Assert( ( ( tempBuffer->geometry.point_count + 1 ) * 2 ) < COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE );
	uint8 * ptr = ( uint8 * )( tempBuffer->point_data + ( tempBuffer->geometry.point_count * 2 ) );
	_write( ptr, vec2, pointA );
	tempBuffer->geometry.point_count++;
}

internal void
addCylinder( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer, vec3 origin, vec3 vector, flo32 radius ) {
	Assert( ( ( tempBuffer->geometry.cylinder_count + 1 ) * 7 ) < COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE );
	uint8 * ptr = ( uint8 * )( tempBuffer->cylinder_data + ( tempBuffer->geometry.cylinder_count * 7 ) );
	_write( ptr, vec3,  origin );
	_write( ptr, vec3,  vector );
	_write( ptr, flo32, radius );
	tempBuffer->geometry.cylinder_count++;
}

internal void
addTorus( COLLISION_GEOMETRY__TEMP_BUFFER * tempBuffer, vec3 origin, vec3 vector, flo32 major_radius, flo32 minor_radius ) {
	Assert( ( ( tempBuffer->geometry.torus_count + 1 ) * 8 ) < COLLISION_GEOMETRY__TEMP_BUFFER__MAX_SIZE );
	uint8 * ptr = ( uint8 * )( tempBuffer->torus_data + ( tempBuffer->geometry.torus_count * 8 ) );
	_write( ptr, vec3,  origin );
	_write( ptr, vec3,  vector );
	_write( ptr, flo32, major_radius );
	_write( ptr, flo32, minor_radius );
	tempBuffer->geometry.torus_count++;
}

//----------
// torus functions
//----------

struct FUNCTION_SOLUTION {
	int32 root_count;
	flo64 root[ 4 ];
};

#define	TWO_PI  6.28318530717958648

internal flo64
N4Step( flo64 x, flo64 a, flo64 b, flo64 c, flo64 d ) { // one Newton step for x^4 + a * x^3 + b * x^2 + c * x + d
	flo64 result = 1e99;
	flo64 fxs    = ( ( 4 * x + 3 * a ) * x + 2 * b ) * x + c; // f'( x )
	if( fxs != 0.0 ) {
		flo64 fx  = ( ( ( x + a ) * x + b ) * x + c ) * x + d; // f( x )
		result    = x - fx / fxs; 
	}
	return result;
} 

#define  SWAP( a, b ) { t = b; b = a; a = t; }
internal void
dblSort3( flo64 & a, flo64 & b, flo64 & c ) { // make: a <= b <= c
	flo64 t;
	if( a > b ) { SWAP( a, b ); } // now a <= b
	if( c < b ) {
		SWAP( b, c ); // now a <= b, b <= c
		if( a > b ) { SWAP( a, b ); } // now a <= b
	}
}

int SolveP3(flo64 *x,flo64 a,flo64 b,flo64 c) {	// solve cubic equation x^3 + a*x^2 + b*x + c
	flo64 a2 = a*a;
    flo64 q  = (a2 - 3*b)/9; 
	flo64 r  = (a*(2*a2-9*b) + 27*c)/54;
    flo64 r2 = r*r;
	flo64 q3 = q*q*q;
	flo64 A,B;
    if(r2<q3) {
        flo64 t=r/sqrt(q3);
		if( t<-1) t=-1;
		if( t> 1) t= 1;
        t=acos(t);
        a/=3; q=-2*sqrt(q);
        x[0]=q*cos(t/3)-a;
        x[1]=q*cos((t+TWO_PI)/3)-a;
        x[2]=q*cos((t-TWO_PI)/3)-a;
        return(3);
    } else {
        A =-pow(fabs(r)+sqrt(r2-q3),1./3); 
		if( r<0 ) A=-A;
		B = A==0? 0 : B=q/A;

		a/=3;
		x[0] =(A+B)-a;
        x[1] =-0.5*(A+B)-a;
        x[2] = 0.5*sqrt(3.)*(A-B);
		if(fabs(x[2])<1e-14) { x[2]=x[1]; return(2); }
        return(1);
    }
}// SolveP3(flo64 *x,flo64 a,flo64 b,flo64 c)

void  CSqrt( flo64 x, flo64 y, flo64 &a, flo64 &b) // returns:  a+i*s = sqrt(x+i*y)
{
	flo64 r  = sqrt(x*x+y*y);
	if( y==0 ) { 
		r = sqrt(r);
		if(x>=0) { a=r; b=0; } else { a=0; b=r; }
	} else {		// y != 0
		a = sqrt(0.5*(x+r));
		b = 0.5*y/a;
	}
}

int   SolveP4Bi(flo64 *x, flo64 b, flo64 d)	// solve equation x^4 + b*x^2 + d = 0
{
	flo64 D = b*b-4*d;
	if( D>=0 ) 
	{
		flo64 sD = sqrt(D);
		flo64 x1 = (-b+sD)/2;
		flo64 x2 = (-b-sD)/2;	// x2 <= x1
		if( x2>=0 )				// 0 <= x2 <= x1, 4 real roots
		{
			flo64 sx1 = sqrt(x1);
			flo64 sx2 = sqrt(x2);
			x[0] = -sx1;
			x[1] =  sx1;
			x[2] = -sx2;
			x[3] =  sx2;
			return 4;
		}
		if( x1 < 0 )				// x2 <= x1 < 0, two pair of imaginary roots
		{
			flo64 sx1 = sqrt(-x1);
			flo64 sx2 = sqrt(-x2);
			x[0] =    0;
			x[1] =  sx1;
			x[2] =    0;
			x[3] =  sx2;
			return 0;
		}
		// now x2 < 0 <= x1 , two real roots and one pair of imginary root
			flo64 sx1 = sqrt( x1);
			flo64 sx2 = sqrt(-x2);
			x[0] = -sx1;
			x[1] =  sx1;
			x[2] =    0;
			x[3] =  sx2;
			return 2;
	} else { // if( D < 0 ), two pair of compex roots
		flo64 sD2 = 0.5*sqrt(-D);
		CSqrt(-0.5*b, sD2, x[0],x[1]);
		CSqrt(-0.5*b,-sD2, x[2],x[3]);
		return 0;
	} // if( D>=0 ) 
} // SolveP4Bi(flo64 *x, flo64 b, flo64 d)	// solve equation x^4 + b*x^2 d

int   SolveP4De(flo64 *x, flo64 b, flo64 c, flo64 d)	// solve equation x^4 + b*x^2 + c*x + d
{
	//if( c==0 ) return SolveP4Bi(x,b,d); // After that, c!=0
	if( fabs(c)<1e-14*(fabs(b)+fabs(d)) ) return SolveP4Bi(x,b,d); // After that, c!=0

	int res3 = SolveP3( x, 2*b, b*b-4*d, -c*c);	// solve resolvent
	// by Viet theorem:  x1*x2*x3=-c*c not equals to 0, so x1!=0, x2!=0, x3!=0
	if( res3>1 )	// 3 real roots, 
	{				
		dblSort3(x[0], x[1], x[2]);	// sort roots to x[0] <= x[1] <= x[2]
		// Note: x[0]*x[1]*x[2]= c*c > 0
		if( x[0] > 0) // all roots are positive
		{
			flo64 sz1 = sqrt(x[0]);
			flo64 sz2 = sqrt(x[1]);
			flo64 sz3 = sqrt(x[2]);
			// Note: sz1*sz2*sz3= -c (and not equal to 0)
			if( c>0 )
			{
				x[0] = (-sz1 -sz2 -sz3)/2;
				x[1] = (-sz1 +sz2 +sz3)/2;
				x[2] = (+sz1 -sz2 +sz3)/2;
				x[3] = (+sz1 +sz2 -sz3)/2;
				return 4;
			}
			// now: c<0
			x[0] = (-sz1 -sz2 +sz3)/2;
			x[1] = (-sz1 +sz2 -sz3)/2;
			x[2] = (+sz1 -sz2 -sz3)/2;
			x[3] = (+sz1 +sz2 +sz3)/2;
			return 4;
		} // if( x[0] > 0) // all roots are positive
		// now x[0] <= x[1] < 0, x[2] > 0
		// two pair of comlex roots
		flo64 sz1 = sqrt(-x[0]);
		flo64 sz2 = sqrt(-x[1]);
		flo64 sz3 = sqrt( x[2]);

		if( c>0 )	// sign = -1
		{
			x[0] = -sz3/2;					
			x[1] = ( sz1 -sz2)/2;		// x[0]i*x[1]
			x[2] =  sz3/2;
			x[3] = (-sz1 -sz2)/2;		// x[2]i*x[3]
			return 0;
		}
		// now: c<0 , sign = +1
		x[0] =   sz3/2;
		x[1] = (-sz1 +sz2)/2;
		x[2] =  -sz3/2;
		x[3] = ( sz1 +sz2)/2;
		return 0;
	} // if( res3>1 )	// 3 real roots, 
	// now resoventa have 1 real and pair of compex roots
	// x[0] - real root, and x[0]>0, 
	// x[1]i*x[2] - complex roots, 
	flo64 sz1 = sqrt(x[0]);
	flo64 szr, szi;
	CSqrt(x[1], x[2], szr, szi);  // (szr+i*szi)^2 = x[1]+i*x[2]
	if( c>0 )	// sign = -1
	{
		x[0] = -sz1/2-szr;			// 1st real root
		x[1] = -sz1/2+szr;			// 2nd real root
		x[2] = sz1/2; 
		x[3] = szi;
		return 2;
	}
	// now: c<0 , sign = +1
	x[0] = sz1/2-szr;			// 1st real root
	x[1] = sz1/2+szr;			// 2nd real root
	x[2] = -sz1/2;
	x[3] = szi;
	return 2;
} // SolveP4De(flo64 *x, flo64 b, flo64 c, flo64 d)	// solve equation x^4 + b*x^2 + c*x + d

internal int32
SolveP4( flo64 * x, flo64 a, flo64 b, flo64 c, flo64 d ) { // solve equation x^4 + a*x^3 + b*x^2 + c*x + d by Dekart-Euler method
	// move to a == 0:
	flo64 d1 = d + 0.25  * a * ( 0.25 * b * a - 3.0 / 64.0 * a * a * a - c );
	flo64 c1 = c + 0.5   * a * ( 0.25 * a * a - b );
	flo64 b1 = b - 0.375 * a * a;
	int32 res = SolveP4De( x, b1, c1, d1 );
	      if( res == 4 ) { x[ 0 ] -= a / 4.0; x[ 1 ] -= a / 4.0; x[ 2 ] -= a / 4.0; x[ 3 ] -= a / 4.0; }
	else if ( res == 2 ) { x[ 0 ] -= a / 4.0; x[ 1 ] -= a / 4.0; x[ 2 ] -= a / 4.0; }
	else                 { x[ 0 ] -= a / 4.0; x[ 2 ] -= a / 4.0; }
	
	// one Newton step for each real root:
	if( res > 0 ) {
		x[ 0 ] = N4Step( x[ 0 ], a, b, c, d );
		x[ 1 ] = N4Step( x[ 1 ], a, b, c, d );
	}
	if( res > 2 ) {
		x[ 2 ] = N4Step( x[ 2 ], a, b, c, d );
		x[ 3 ] = N4Step( x[ 3 ], a, b, c, d );
	}
	return res;
}

internal FUNCTION_SOLUTION
solveQuarticEquation( flo32 B, flo32 C, flo32 D, flo32 E ) {
	flo64 roots[ 4 ] = {};
	int32 root_count = SolveP4( roots, B, C, D, E );
	
	FUNCTION_SOLUTION quartic = { root_count, roots[0], roots[1], roots[2], roots[3] };
	return quartic;
}

internal COLLISION_3D
collisionRT( vec3 ray_origin, vec3 ray_vector, vec3 torus_center, vec3 torus_normal, flo32 torus_majorRadius, flo32 torus_minorRadius ) {
	COLLISION_3D result = {};
	
	vec3 vectorA = torus_normal;
	vec3 vectorB = getNormal( ray_vector );
	vec3 vectorC = ray_origin - torus_center;
	
	flo32 radA = torus_majorRadius;
	flo32 radB = torus_minorRadius;
	flo32 ray_vectorLength = getLength( ray_vector );
	
	flo32 dot_AB = dot( vectorA, vectorB );
	flo32 dot_AC = dot( vectorA, vectorC );
	flo32 dot_BC = dot( vectorB, vectorC );
	flo32 dot_CC = dot( vectorC, vectorC );
	
	flo32 radA_sq = squareValue( radA );
	
	flo32 a = 1.0f - squareValue( dot_AB );
	flo32 b = 2.0f * ( dot_BC - dot_AC * dot_AB );
	flo32 c = dot_CC - dot_AC * dot_AC;
	flo32 d = dot_CC + radA_sq - radB * radB;
	
	flo32 radA_sq_4x = 4.0f * radA_sq;
	flo32 B = 4.0f * dot_BC;
	flo32 C = 2.0f * d + 0.25f * B * B - radA_sq_4x * a;
	flo32 D = B * d - radA_sq_4x * b;
	flo32 E = d * d - radA_sq_4x * c;
	
	FUNCTION_SOLUTION quartic = solveQuarticEquation( B, C, D, E );
	
	if( quartic.root_count > 0 ) {
		flo32 tMin = 1000000.0f;
		for( int32 root_index = 0; root_index < quartic.root_count; root_index++ ) {
			flo32 root_current = ( flo32 )quartic.root[ root_index ];
			flo32 t            = root_current / ray_vectorLength;
			if( ( t > -COLLISION__COLLISION_EPSILON ) && ( t < tMin ) ) {
				tMin = t;
			}
		}
		if( ( tMin > -COLLISION__COLLISION_EPSILON ) && ( tMin < ( 1.0f + COLLISION__COLLISION_EPSILON ) ) ) {
			vec3 collision_point = ray_origin + ray_vector * tMin;
			vec3 vectorD = collision_point - torus_center;
			vec3 vectorE = getNormal( vectorD - dot( vectorA, vectorD ) * vectorA );
			vec3 normal  = getNormal( vectorD - vectorE * radA );
			if( dot( normal, ray_vector ) < -COLLISION__COLLISION_EPSILON ) {
				result.isValid = true;
				result.t       = clamp01( tMin );
				result.normal  = normal;
			}
		}
	}
	return result;
}

//----------
// cylinder functions
//----------

internal boo32
isInCylinder( vec3 point, cylinder cylinder ) {
	boo32 result = false;
	
	vec3 vectorA = cylinder.vector;
	vec3 vectorB = point - cylinder.origin;
	
	flo32 dotAA = dot( vectorA, vectorA );
	flo32 dotAB = dot( vectorA, vectorB );
	flo32 dotBB = dot( vectorB, vectorB );
	
	flo32 distSq = dotBB - ( dotAB * dotAB ) / dotAA;
	if( distSq <= ( cylinder.radius * cylinder.radius ) ) {
		flo32 t = dotAB / dotAA;
		if( ( t >= 0.0f ) && ( t <= 1.0f ) ) {
			result = true;
		}
	}
	return result;
}

internal COLLISION_3D
collision_cylinder( vec3 ray_origin, vec3 ray_vector, vec3 cylinder_origin, vec3 cylinder_vector, flo32 cylinder_radius ) {
	COLLISION_3D result = {};
	
	flo32 s      =  0.0f;
	flo32 distSq = -1.0f;
	
	{ // get dist sq
		vec3 U = ray_vector;
		vec3 V = cylinder_vector;
		vec3 W = ray_origin - cylinder_origin;
		
		flo32 a = dot( U, U );
		flo32 b = dot( U, V );
		flo32 c = dot( V, V );
		flo32 d = dot( U, W );
		flo32 e = dot( V, W );
		
		flo32 denom = a * c - b * b;
		if( absValue( denom ) > COLLISION__COLLISION_EPSILON ) {
			      s = ( b * e - c * d ) / denom;
			flo32 t = ( a * e - b * d ) / denom;
			
			vec3 pointA = ray_origin + ray_vector * s;
			vec3 pointB = cylinder_origin + cylinder_vector * t;
			distSq = getLengthSq( pointB - pointA );
		}
	}
	
	flo32 bSq = ( cylinder_radius * cylinder_radius ) - distSq;
	if( bSq > COLLISION__COLLISION_EPSILON ) {
		flo32 denom = 1.0f - ( squareValue( dot( ray_vector, cylinder_vector ) ) / ( getLengthSq( ray_vector ) * getLengthSq( cylinder_vector ) ) );
		if( absValue( denom ) > COLLISION__COLLISION_EPSILON ) {
			flo32 s0 = s - sqrtf( bSq / ( denom * getLengthSq( ray_vector ) ) );
			if( ( s0 > -COLLISION__COLLISION_EPSILON ) && ( s0 < ( 1.0f + COLLISION__COLLISION_EPSILON ) ) ) {
				vec3  intersect_vector = ( ray_origin + ray_vector * s0 ) - cylinder_origin;
				flo32 intersect_ratio  = ( dot( intersect_vector, cylinder_vector ) / dot( cylinder_vector, cylinder_vector ) );
				if( ( intersect_ratio > -COLLISION__COLLISION_EPSILON ) && ( intersect_ratio < 1.0f + COLLISION__COLLISION_EPSILON ) ) {
					vec3 normal = getNormal( intersect_vector - cylinder_vector * intersect_ratio );
					if( dot( normal, ray_vector ) < -COLLISION__COLLISION_EPSILON ) {
						result.isValid = true;
						result.t       = clamp01( s0 );
						result.normal  = normal;
					}
				}
			}
		}
	}
	return result;
}

internal COLLISION_3D
collision_cylinder( vec3 ray_origin, vec3 ray_vector, cylinder cylinder ) {
	COLLISION_3D result = collision_cylinder( ray_origin, ray_vector, cylinder.origin, cylinder.vector, cylinder.radius );
	return result;
}

internal COLLISION_3D
collision_cylinder( vec3 ray_vector, vec3 ray_origin, vec3 * vertex, flo32 cylinder_radius ) {
	vec3 cylinder_origin = vertex[ 0 ];
	vec3 cylinder_vector = vertex[ 1 ] - vertex[ 0 ];
	COLLISION_3D result = collision_cylinder( ray_vector, ray_origin, cylinder_origin, cylinder_vector, cylinder_radius );
	return result;
}


//----------
// sphere functions
//----------

internal COLLISION_3D
collisionRS( vec3 ray_origin, vec3 ray_vector, vec3 sphere_origin, flo32 sphere_radius ) {
	COLLISION_3D result = {};
	vec3 sphere_vector = sphere_origin - ray_origin;
	
	flo32 a = dot( ray_vector, sphere_vector );
	flo32 b = dot( ray_vector, ray_vector );
	flo32   distSq = dot( sphere_vector, sphere_vector ) - ( ( a * a ) / b );
	flo32 radiusSq = sphere_radius * sphere_radius;
	if( distSq < radiusSq ) {
		flo32 t  = a / b;
		flo32 t0 = t - ( sqrtf( radiusSq - distSq ) / getLength( ray_vector ) );
		if( ( t0 > -COLLISION__COLLISION_EPSILON ) && ( t0 < ( 1.0f + COLLISION__COLLISION_EPSILON ) ) ) {
			vec3 normal = getNormal( ray_vector * t0 - sphere_vector );
			if( dot( normal, ray_vector ) < -COLLISION__COLLISION_EPSILON ) {
				result.isValid = true;
				result.t       = clamp01( t0 );
				result.normal  = normal;
			}
		}
	}
	return result;
}

internal COLLISION_3D
collisionRS( vec3 ray_origin, vec3 ray_vector, sphere bound ) {
	COLLISION_3D result = collisionRS( ray_origin, ray_vector, bound.origin, bound.radius );
	return result;
}

//----------
// circle functions
//----------

internal COLLISION_2D
collision_circle( vec2 ray_origin, vec2 ray_vector, vec2 * center, int32 center_count, flo32 radius ) {
	COLLISION_2D result = { false, 1000000.0f, };
	COLLISION_2D collision = {};
	
	flo32 radiusSq = radius * radius;
	for( int32 center_index = 0; center_index < center_count; center_index++ ) {
		vec2 vectorA = center[ center_index ] - ray_origin;
		vec2 vectorB = ray_vector;
		
		flo32 dotAA = dot( vectorA, vectorA );
		flo32 dotAB = dot( vectorA, vectorB );
		flo32 dotBB = dot( vectorB, vectorB );
		
		flo32 distSq = dotAA - ( ( dotAB * dotAB ) / dotBB );
		if( distSq < radiusSq ) {
			flo32 t = ( dotAB / dotBB ) - sqrtf( ( radiusSq - distSq ) / dotBB );
			
			flo32 t0 = t * getLength( vectorB );
			if( ( t0 > -0.001f ) && ( t0 < 0.0f ) ) {
				t = 0.0f;
			}
			
			if( ( t >= 0.0f ) && ( t < result.t ) ) {
				vec2 normal = getNormal( vectorB * t - vectorA );
				if( dot( normal, vectorB ) < 0.0f ) {
					result.isValid = true;
					result.t       = t;
					result.normal  = normal;
				}
			}
		}
	}
	return result;
}

//----------
// quad functions
//----------

internal COLLISION_3D
collision_quad( vec3 ray_origin, vec3 ray_vector, vec3 * vertex, flo32 offset ) {
	vec3 normal = getNormal( cross( vertex[ 1 ] - vertex[ 0 ], vertex[ 2 ] - vertex[ 0 ] ) );
	vec3 origin = ray_origin - normal * offset;
	vec3 vector = ray_vector;
	
	COLLISION_3D result = {};
	flo32 t0 = dot( vector, normal );
	if( t0 < 0.0f ) {
		flo32 t = dot( ( vertex[ 0 ] - origin ), normal ) / t0;
		if( ( t >= -COLLISION__COLLISION_EPSILON ) && ( t < 0.0f ) ) { t = 0.0f; }
		if( t >= 0.0f ) {
			vec3 point = origin + vector * t;
			
			flo32 dot0 = dot( point - vertex[ 0 ], cross( normal, vertex[ 1 ] - vertex[ 0 ] ) );
			flo32 dot1 = dot( point - vertex[ 1 ], cross( normal, vertex[ 2 ] - vertex[ 1 ] ) );
			flo32 dot2 = dot( point - vertex[ 2 ], cross( normal, vertex[ 3 ] - vertex[ 2 ] ) );
			flo32 dot3 = dot( point - vertex[ 3 ], cross( normal, vertex[ 0 ] - vertex[ 3 ] ) );
			
			boo32 isValid = ( ( dot0 >= 0.0f ) && ( dot1 >= 0.0f ) && ( dot2 >= 0.0f ) && ( dot3 >= 0.0f ) );
			
			if( isValid ) {
				result.isValid = true;
				result.t       = t;
				result.normal  = normal;
			}
		}
	}
	return result;
}

//----------
// triangle functions
//----------

internal COLLISION_3D
collision_tri( vec3 ray_origin, vec3 ray_vector, vec3 * vertex, flo32 offset ) {
	vec3 normal = getNormal( cross( vertex[ 1 ] - vertex[ 0 ], vertex[ 2 ] - vertex[ 0 ] ) );
	vec3 origin = ray_origin - normal * offset;
	vec3 vector = ray_vector;
	
	COLLISION_3D result = {};
	flo32 t0 = dot( vector, normal );
	if( t0 < -0.000001f ) {
		flo32 t = dot( ( vertex[ 0 ] - origin ), normal ) / t0;
		if( ( t >= -COLLISION__COLLISION_EPSILON ) && ( t < 0.0f ) ) { t = 0.0f; }
		if( t >= 0.0f ) {
			vec3 point = origin + vector * t;
			
			flo32 dot0 = dot( point - vertex[ 0 ], cross( normal, vertex[ 1 ] - vertex[ 0 ] ) );
			flo32 dot1 = dot( point - vertex[ 1 ], cross( normal, vertex[ 2 ] - vertex[ 1 ] ) );
			flo32 dot2 = dot( point - vertex[ 2 ], cross( normal, vertex[ 0 ] - vertex[ 2 ] ) );
			
			boo32 isValid = ( ( dot0 >= 0.0f ) && ( dot1 >= 0.0f ) && ( dot2 >= 0.0f ) );
			
			if( isValid ) {
				result.isValid = true;
				result.t       = t;
				result.normal  = normal;
			}
		}
	}
	return result;
}

//----------
// ray functions
//----------

internal COLLISION_3D
collision_ray( vec3 ray_origin, vec3 ray_vector, COLLISION_GEOMETRY * collisionGeometry ) {
	COLLISION_3D result = {};
	
	flo32 tMin = 1000000.0f;
	
	flo32 * data = ( flo32 * )( collisionGeometry + 1 );
	for( int32 quad_index = 0; quad_index < collisionGeometry->quad_count; quad_index++ ) {
		COLLISION_3D collision = collision_quad( ray_origin, ray_vector, ( vec3 * )data, 0.0f );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 12;
	}
	for( int32 tri_index = 0; tri_index < collisionGeometry->tri_count; tri_index++ ) {
		COLLISION_3D collision = collision_tri( ray_origin, ray_vector, ( vec3 * )data, 0.0f );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 9;
	}
	
	data += ( 6 * collisionGeometry->edge_count  );
	data += ( 3 * collisionGeometry->point_count );
	
	for( int32 cylinder_index = 0; cylinder_index < collisionGeometry->cylinder_count; cylinder_index++ ) {
		COLLISION_3D collision = collision_cylinder( ray_origin, ray_vector, ( vec3 * )data, data[ 6 ] );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 7;
	}
	for( int32 torus_index = 0; torus_index < collisionGeometry->torus_count; torus_index++ ) {
		COLLISION_3D collision = collisionRT( ray_origin, ray_vector, ( ( vec3 * )data )[ 0 ], ( ( vec3 * )data )[ 1 ], data[ 6 ], data[ 7 ] );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 8;
	}
	
	return result;
}

//----------
// edge functions
//----------

internal COLLISION_2D
collision_edge( vec2 ray_origin, vec2 ray_vector, vec2 * vertex, int32 edge_count, flo32 offset ) {
	COLLISION_2D result = { false, 1000000.0f, };
	
	int32 vertex_count = edge_count * 2;
	for( int32 vertex_index = 0; vertex_index < vertex_count; vertex_index += 2 ) {
		vec2 vertex0 = vertex[ vertex_index     ];
		vec2 vertex1 = vertex[ vertex_index + 1 ];
		
		vec2 vector  = vertex1 - vertex0;
		vec2 normal  = getNormal( getPerp( vector ) );
		
		if( dot( ray_vector, normal ) < 0.0f ) {
			vec2 p = ray_origin;
			vec2 r = ray_vector;
			
			vec2 q = vertex0 + normal * offset;
			vec2 s = vector;
			
			vec2 w = q - p;
			
			flo32 denom = ( ( r.x * s.y ) - ( r.y * s.x ) );
			if( denom != 0.0f ) {
				flo32 t = ( ( w.x * s.y ) - ( w.y * s.x ) ) / denom;
				flo32 u = ( ( w.x * r.y ) - ( w.y * r.x ) ) / denom;
				
				flo32 t0 = t * getLength( r );
				if( ( t0 > -0.001f ) && ( t0 < 0.0f ) ) {
					t = 0.0f;
				}
				
				if( ( u >= 0.0f ) && ( u <= 1.0f ) && ( t >= 0.0f ) && ( t < result.t ) ) {
					result.isValid = true;
					result.t       = t;
					result.normal  = normal;
				}
			}
		}
	}
	
	return result;
}

//----------
// minkowski functions
//----------

internal COLLISION_3D
collision_sphere( vec3 ray_origin, vec3 ray_vector, flo32 sphere_radius, COLLISION_GEOMETRY * collisionGeometry ) {
	COLLISION_3D result = {};
	
	flo32 tMin = 1.0f;
	
	flo32 * data = ( flo32 * )( collisionGeometry + 1 );
	for( int32 quad_index = 0; quad_index < collisionGeometry->quad_count; quad_index++ ) {
		COLLISION_3D collision = collision_quad( ray_origin, ray_vector, ( vec3 * )data, sphere_radius );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 12;
	}
	for( int32 tri_index = 0; tri_index < collisionGeometry->tri_count; tri_index++ ) {
		COLLISION_3D collision = collision_tri( ray_origin, ray_vector, ( vec3 * )data, sphere_radius );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 9;
	}
	for( int32 edge_index = 0; edge_index < collisionGeometry->edge_count; edge_index++ ) {
		COLLISION_3D collision = collision_cylinder( ray_origin, ray_vector, ( vec3 * )data, sphere_radius );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 6;
	}
	for( int32 point_index = 0; point_index < collisionGeometry->point_count; point_index++ ) {
		COLLISION_3D collision = collisionRS( ray_origin, ray_vector, ( ( vec3 * )data )[ 0 ], sphere_radius );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 3;
	}
	for( int32 cylinder_index = 0; cylinder_index < collisionGeometry->cylinder_count; cylinder_index++ ) {
		COLLISION_3D collision = collision_cylinder( ray_origin, ray_vector, ( vec3 * )data, sphere_radius + data[ 6 ] );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 7;
	}
	for( int32 torus_index = 0; torus_index < collisionGeometry->torus_count; torus_index++ ) {
		COLLISION_3D collision = collisionRT( ray_origin, ray_vector, ( ( vec3 * )data )[ 0 ], ( ( vec3 * )data )[ 1 ], data[ 6 ], data[ 7 ] + sphere_radius );
		if( ( collision.isValid ) && ( collision.t < tMin ) ) {
			result = collision;
			tMin   = collision.t;
		}
		data += 8;
	}
	
	return result;
}

internal COLLISION_2D
collision_circle( vec2 ray_origin, vec2 ray_vector, flo32 circle_radius, COLLISION_GEOMETRY * collisionGeometry ) {
	COLLISION_2D result = { false, 1000000.0f, };
	COLLISION_2D collision = {};
	
	flo32 * data = ( flo32 * )( collisionGeometry + 1 );
	
	Assert( collisionGeometry->quad_count == 0 );
	Assert( collisionGeometry->tri_count  == 0 );
	
	collision = collision_edge( ray_origin, ray_vector, ( vec2 * )data, collisionGeometry->edge_count, circle_radius );
	if( ( collision.isValid ) && ( collision.t < result.t ) ) {
		result = collision;
	}
	data += ( 4 * collisionGeometry->edge_count );
	
	collision = collision_circle( ray_origin, ray_vector, ( vec2 * )data, collisionGeometry->point_count, circle_radius );
	if( ( collision.isValid ) && ( collision.t < result.t ) ) {
		result = collision;
	}
	data += ( 2 * collisionGeometry->point_count );
	
	Assert( collisionGeometry->cylinder_count == 0 );
	Assert( collisionGeometry->torus_count    == 0 );
	
	return result;
}

#endif	// STD_INCLUDE_FUNC