
#ifdef	STD_INCLUDE_DECL

//----------
// debug declarations
//----------

internal void CONSOLE_STRING( const char * string );

enum DEBUG_MESH_ID {
	debugMeshID_null,
	
	debugMeshID_centerPanel,
	debugMeshID_cube,
	debugMeshID_sphere,
	debugMeshID_cylinder,
	debugMeshID_torus,
	
	debugMeshID_count,
};

struct DEBUG_MESH {
	MESH_DRAW_TYPE drawType;
	int32 vertex_index;
	int32 vertex_count;
};

struct DEBUG_SYSTEM {
	RENDER_SYSTEM_2D * renderSystem2D;
	RENDER_SYSTEM_3D * renderSystem3D;
	
	DEBUG_MESH        debugMesh[ debugMeshID_count ];
};

#if DEBUG_BUILD
#define DISPLAY_VALUE( type, name ) {\
													char __debug_string[ 2048 ] = {};\
													serialize_##type( __debug_string, #name, name );\
													DISPLAY_STRING( __debug_string );\
												}
#define CONSOLE_VALUE( type, name ) {\
													char __debug_string[ 2048 ] = {};\
													serialize_##type( __debug_string, #name, name );\
													strcat( __debug_string, "\n" );\
													OutputDebugString( __debug_string );\
												}

#else
#define DISPLAY_STRING( str )
#define DISPLAY_VALUE( type, name )
#define CONSOLE_STRING( str )

internal void PROFILE_string( char * string ) {}
internal void PROFILE_tooltip( char * string, vec2 position ) {}

internal void debug_cube( cube bound, vec4 color ) {}
internal void debug_sphere( vec3 position, flo32 radius, vec4 color ) {}
internal void debug_cylinder( vec3 position, flo32 height, flo32 radius, vec4 color ) {}
internal void debug_cylinder( cylinder cylinder, vec4 color ) {}
internal void debug_torus( vec3 position, vec4 color ) {}
internal void debug_line( vec3 pointA, vec3 pointB, vec4 color ) {}

#endif

#endif	// STD_INCLUDE_DECL
#ifdef	STD_INCLUDE_FUNC

//----------
// debug functions
//----------

#if DEBUG_BUILD
#define DEBUG_SYSTEM__MAX_VERTEX_COUNT		( 262144 )
#define DEBUG_SYSTEM__MAX_MEMORY_SIZE    _MB( 2 )
#define DEBUG_SYSTEM__SIZE_IN_BYTES  		( sizeof( RENDER_SYSTEM_2D ) + sizeof( RENDER_SYSTEM_3D ) + DEBUG_SYSTEM__MAX_MEMORY_SIZE )

DEBUG_SYSTEM globalVar_debugSystem = {};
vec2         globalVar_atPos = {};
flo32        globalVar_advanceHeight = 0.0f;

internal DEBUG_MESH
DebugMesh( MESH_DRAW_TYPE drawType, int32 vertex_index, int32 vertex_count ) {
	DEBUG_MESH result = { drawType, vertex_index, vertex_count };
	return result;
}

internal void
CONSOLE_STRING( const char * string ) {
	OutputDebugString( string );
	OutputDebugString( "\n" );
}

internal void
DEBUG_addTask( uint32 texture_bufferID, rect bound, vec4 color ) {
	RENDER_SYSTEM_2D * renderSystem = globalVar_debugSystem.renderSystem2D;
	
	beginTask( renderSystem, meshDrawType_triangleFan, texture_bufferID, color );
	addQuad( renderSystem, bound, Rect( 0.0f, 0.0f, 1.0f, 1.0f ) );
	endTask( renderSystem );
}

internal void
DEBUG_drawString( char * string, vec2 position, vec4 color ) {
	RENDER_SYSTEM_2D * renderSystem = globalVar_debugSystem.renderSystem2D;
	
	RENDER_MONOSPACE_FONT * renderFont = getMonoFont( renderSystem->assetSystem, assetID_monoFont_default );
	MONOSPACE_FONT * font = &renderFont->font;
	
	beginTask( renderSystem, meshDrawType_triangles, renderFont->texture_bufferID, color );
	
	int32 char_count = ( int32 )strlen( string );
	
	vec2 atPos = position;
	for( int32 counter = 0; counter < char_count; counter++ ) {
		char c = string[ counter ];
		if( c != ' ' ) {
			int32 char_index = c - FONT_START_CHAR;
			MONOSPACE_FONT_CHAR _char = font->alphabet[ char_index ];
			
			flo32 left   = atPos.x + _char.offset.x;
			flo32 right  = left    + _char.dim.x;
			flo32 top    = atPos.y + _char.offset.y;
			flo32 bottom = top     - _char.dim.y;
			rect bound = Rect( left, bottom, right, top );
			
			addQuad( renderSystem, bound, rectMM( _char.texCoord_min, _char.texCoord_max ) );
		}
		atPos.x += font->advanceWidth;
	}
	endTask( renderSystem );
}

internal void
DEBUG_drawRect( rect bound, vec4 color ) {
	uint32 texture_bufferID = getTexture( globalVar_debugSystem.renderSystem2D->assetSystem, assetID_texture_whiteTexture ).bufferID;
	DEBUG_addTask( texture_bufferID, bound, color );
}

internal void
DISPLAY_STRING( char * string ) {
	DEBUG_drawString( string, globalVar_atPos, COLOR_BLACK );
	globalVar_atPos.y -= globalVar_advanceHeight;
}

internal void
DISPLAY_STRING( char * string, vec2 position ) {
	DEBUG_drawString( string, position, COLOR_BLACK );
}

internal void
PROFILE_string( char * string ) {
	DEBUG_drawString( string, globalVar_atPos, COLOR_WHITE );
	globalVar_atPos.y -= globalVar_advanceHeight;
}

internal void
PROFILE_tooltip( char * string, vec2 position ) {
	DEBUG_drawString( string, position, COLOR_WHITE );
}

internal void
DEBUG_genMesh_cube( VERTEX_LIST * vertexList ) {
	int32 vertex_count = 24;
	int32 vertex_index = vertexList->vertex_count;
	
	vec3  point[] = {
		Vec3( -1.0f, -1.0f, -1.0f ), Vec3(  1.0f, -1.0f, -1.0f ),
		Vec3(  1.0f, -1.0f,  1.0f ), Vec3( -1.0f, -1.0f,  1.0f ),
		Vec3( -1.0f,  1.0f, -1.0f ), Vec3(  1.0f,  1.0f, -1.0f ),
		Vec3(  1.0f,  1.0f,  1.0f ), Vec3( -1.0f,  1.0f,  1.0f ),
	};
	
	addVertex( vertexList, point[ 0 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 3 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 3 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 7 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 7 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 4 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 4 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 0 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 2 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 1 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 1 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 5 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 5 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 6 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 6 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 2 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 0 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 1 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 3 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 2 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 7 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 6 ], Vec2( 0.0f, 0.0f ) );
	addVertex( vertexList, point[ 4 ], Vec2( 0.0f, 0.0f ) );   addVertex( vertexList, point[ 5 ], Vec2( 0.0f, 0.0f ) );
	
	globalVar_debugSystem.debugMesh[ debugMeshID_cube ] = DebugMesh( meshDrawType_lines, vertex_index, vertex_count );
}

internal void
debug_cube( cube bound, vec4 color ) {
	RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
	
	DEBUG_MESH mesh = globalVar_debugSystem.debugMesh[ debugMeshID_cube ];
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, assetID_texture_whiteTexture ).bufferID;
	addRenderObject( renderSystem, meshDrawType_lines, mesh.vertex_index, mesh.vertex_count, getCenter( bound ), getHalfDim( bound ), Quat( Vec3( 1.0f, 0.0f, 0.0f ), 0.0f ), texture_bufferID, color );
}

internal void
DEBUG_genMesh_sphere( VERTEX_LIST * vertexList ) {
	int32 vertex_count = 480;
	int32 vertex_index = vertexList->vertex_count;
	
	flo32 sin_lookup[ 5 ] = { 0.0f, sinf( PI * 0.125f ), sinf( PI * 0.25f ), sinf( PI * 0.375f ), 1.0f };
	flo32 cos_lookup[ 5 ] = { 1.0f, cosf( PI * 0.125f ), cosf( PI * 0.25f ), cosf( PI * 0.375f ), 0.0f };
	
	for( int32 iter = 0; iter < 4; iter++ ) {
		vec2 vertexA  = Vec2(  cos_lookup[ iter     ], -sin_lookup[ iter     ] );
		vec2 vertexA0 = Vec2(  cos_lookup[ iter + 1 ], -sin_lookup[ iter + 1 ] );
		vec2 vertexB  = Vec2( -sin_lookup[ iter     ], -cos_lookup[ iter     ] );
		vec2 vertexB0 = Vec2( -sin_lookup[ iter + 1 ], -cos_lookup[ iter + 1 ] );
		vec2 vertexC  = Vec2( -cos_lookup[ iter     ],  sin_lookup[ iter     ] );
		vec2 vertexC0 = Vec2( -cos_lookup[ iter + 1 ],  sin_lookup[ iter + 1 ] );
		vec2 vertexD  = Vec2(  sin_lookup[ iter     ],  cos_lookup[ iter     ] );
		vec2 vertexD0 = Vec2(  sin_lookup[ iter + 1 ],  cos_lookup[ iter + 1 ] );
		
		for( int32 iterB = 0; iterB < 4; iterB++ ) {
			vec2 adjVertexA  = vertexA  * cos_lookup[ iterB     ];
			vec2 adjVertexA0 = vertexA0 * cos_lookup[ iterB     ];
			vec2 adjVertexB  = vertexA  * cos_lookup[ iterB + 1 ];
			vec2 adjVertexC  = vertexB  * cos_lookup[ iterB     ];
			vec2 adjVertexC0 = vertexB0 * cos_lookup[ iterB     ];
			vec2 adjVertexD  = vertexB  * cos_lookup[ iterB + 1 ];
			vec2 adjVertexE  = vertexC  * cos_lookup[ iterB     ];
			vec2 adjVertexE0 = vertexC0 * cos_lookup[ iterB     ];
			vec2 adjVertexF  = vertexC  * cos_lookup[ iterB + 1 ];
			vec2 adjVertexG  = vertexD  * cos_lookup[ iterB     ];
			vec2 adjVertexG0 = vertexD0 * cos_lookup[ iterB     ];
			vec2 adjVertexH  = vertexD  * cos_lookup[ iterB + 1 ];
			
			addVertex( vertexList, Vec3( adjVertexA.x,  sin_lookup[ iterB     ], adjVertexA.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexB.x,  sin_lookup[ iterB + 1 ], adjVertexB.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexA.x, -sin_lookup[ iterB     ], adjVertexA.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexB.x, -sin_lookup[ iterB + 1 ], adjVertexB.y ), Vec2( 0.0f, 0.0f ) );
			
			addVertex( vertexList, Vec3( adjVertexC.x,  sin_lookup[ iterB     ], adjVertexC.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexD.x,  sin_lookup[ iterB + 1 ], adjVertexD.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexC.x, -sin_lookup[ iterB     ], adjVertexC.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexD.x, -sin_lookup[ iterB + 1 ], adjVertexD.y ), Vec2( 0.0f, 0.0f ) );
			
			addVertex( vertexList, Vec3( adjVertexE.x,  sin_lookup[ iterB     ], adjVertexE.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexF.x,  sin_lookup[ iterB + 1 ], adjVertexF.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexE.x, -sin_lookup[ iterB     ], adjVertexE.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexF.x, -sin_lookup[ iterB + 1 ], adjVertexF.y ), Vec2( 0.0f, 0.0f ) );
			
			addVertex( vertexList, Vec3( adjVertexG.x,  sin_lookup[ iterB     ], adjVertexG.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexH.x,  sin_lookup[ iterB + 1 ], adjVertexH.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexG.x, -sin_lookup[ iterB     ], adjVertexG.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexH.x, -sin_lookup[ iterB + 1 ], adjVertexH.y ), Vec2( 0.0f, 0.0f ) );
			
			addVertex( vertexList, Vec3( adjVertexA.x,   sin_lookup[ iterB ], adjVertexA.y  ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexA0.x,  sin_lookup[ iterB ], adjVertexA0.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexC.x,   sin_lookup[ iterB ], adjVertexC.y  ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexC0.x,  sin_lookup[ iterB ], adjVertexC0.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexE.x,   sin_lookup[ iterB ], adjVertexE.y  ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexE0.x,  sin_lookup[ iterB ], adjVertexE0.y ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexG.x,   sin_lookup[ iterB ], adjVertexG.y  ), Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, Vec3( adjVertexG0.x,  sin_lookup[ iterB ], adjVertexG0.y ), Vec2( 0.0f, 0.0f ) );
			
			if( iterB > 0 ) {
				addVertex( vertexList, Vec3( adjVertexA.x,  -sin_lookup[ iterB ], adjVertexA.y  ), Vec2( 0.0f, 0.0f ) );
				addVertex( vertexList, Vec3( adjVertexA0.x, -sin_lookup[ iterB ], adjVertexA0.y ), Vec2( 0.0f, 0.0f ) );
				addVertex( vertexList, Vec3( adjVertexC.x,  -sin_lookup[ iterB ], adjVertexC.y  ), Vec2( 0.0f, 0.0f ) );
				addVertex( vertexList, Vec3( adjVertexC0.x, -sin_lookup[ iterB ], adjVertexC0.y ), Vec2( 0.0f, 0.0f ) );
				addVertex( vertexList, Vec3( adjVertexE.x,  -sin_lookup[ iterB ], adjVertexE.y  ), Vec2( 0.0f, 0.0f ) );
				addVertex( vertexList, Vec3( adjVertexE0.x, -sin_lookup[ iterB ], adjVertexE0.y ), Vec2( 0.0f, 0.0f ) );
				addVertex( vertexList, Vec3( adjVertexG.x,  -sin_lookup[ iterB ], adjVertexG.y  ), Vec2( 0.0f, 0.0f ) );
				addVertex( vertexList, Vec3( adjVertexG0.x, -sin_lookup[ iterB ], adjVertexG0.y ), Vec2( 0.0f, 0.0f ) );
			}
		}
	}
	
	globalVar_debugSystem.debugMesh[ debugMeshID_sphere ] = DebugMesh( meshDrawType_lines, vertex_index, vertex_count );
}

internal void
debug_sphere( vec3 position, flo32 radius, vec4 color ) {
	RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
	
	DEBUG_MESH mesh = globalVar_debugSystem.debugMesh[ debugMeshID_sphere ];
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, assetID_texture_whiteTexture ).bufferID;
	addRenderObject( renderSystem, meshDrawType_lines, mesh.vertex_index, mesh.vertex_count, position, Vec3( radius, radius, radius ), Quat( Vec3( 1.0f, 0.0f, 0.0f ), 0.0f ), texture_bufferID, color );
}

internal void
DEBUG_genMesh_cylinder( VERTEX_LIST * vertexList ) {
	int32 vertex_count = 96;
	int32 vertex_index = vertexList->vertex_count;
	
	flo32 sin_lookup[ 5 ] = { 0.0f, sinf( PI * 0.125f ), sinf( PI * 0.25f ), sinf( PI * 0.375f ), 1.0f };
	flo32 cos_lookup[ 5 ] = { 1.0f, cosf( PI * 0.125f ), cosf( PI * 0.25f ), cosf( PI * 0.375f ), 0.0f };
	
	for( int32 iter = 0; iter < 4; iter++ ) {
		flo32 co = cos_lookup[ iter ];
		flo32 si = sin_lookup[ iter ];
		
		vec2 vertexA = Vec2(  co, -si );
		vec2 vertexB = Vec2( -si, -co );
		vec2 vertexC = Vec2( -co,  si );
		vec2 vertexD = Vec2(  si,  co );
		
		addVertex( vertexList, Vec3( vertexA.x, 0.0f, vertexA.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexA.x, 1.0f, vertexA.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexB.x, 0.0f, vertexB.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexB.x, 1.0f, vertexB.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexC.x, 0.0f, vertexC.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexC.x, 1.0f, vertexC.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexD.x, 0.0f, vertexD.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexD.x, 1.0f, vertexD.y ), Vec2( 0.0f, 0.0f ) );
	}
	
	for( int32 iter = 0; iter < 4; iter++ ) {
		flo32 co = cos_lookup[ iter ];
		flo32 si = sin_lookup[ iter ];
		
		flo32 co0 = cos_lookup[ iter + 1 ];
		flo32 si0 = sin_lookup[ iter + 1 ];
		
		vec2 vertexA = Vec2(   co,  -si );
		vec2 vertexB = Vec2(  -si,  -co );
		vec2 vertexC = Vec2(  -co,   si );
		vec2 vertexD = Vec2(   si,   co );
		vec2 vertexA0 = Vec2(  co0, -si0 );
		vec2 vertexB0 = Vec2( -si0, -co0 );
		vec2 vertexC0 = Vec2( -co0,  si0 );
		vec2 vertexD0 = Vec2(  si0,  co0 );
		
		addVertex( vertexList, Vec3(  vertexA.x, 0.0f,  vertexA.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexA0.x, 0.0f, vertexA0.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3(  vertexA.x, 1.0f,  vertexA.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexA0.x, 1.0f, vertexA0.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3(  vertexB.x, 0.0f,  vertexB.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexB0.x, 0.0f, vertexB0.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3(  vertexB.x, 1.0f,  vertexB.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexB0.x, 1.0f, vertexB0.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3(  vertexC.x, 0.0f,  vertexC.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexC0.x, 0.0f, vertexC0.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3(  vertexC.x, 1.0f,  vertexC.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexC0.x, 1.0f, vertexC0.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3(  vertexD.x, 0.0f,  vertexD.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexD0.x, 0.0f, vertexD0.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3(  vertexD.x, 1.0f,  vertexD.y ), Vec2( 0.0f, 0.0f ) );
		addVertex( vertexList, Vec3( vertexD0.x, 1.0f, vertexD0.y ), Vec2( 0.0f, 0.0f ) );		
	}
	
	globalVar_debugSystem.debugMesh[ debugMeshID_cylinder ] = DebugMesh( meshDrawType_lines, vertex_index, vertex_count );
}

internal void
debug_cylinder( vec3 position, flo32 height, flo32 radius, vec4 color ) {
	RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
	
	DEBUG_MESH mesh = globalVar_debugSystem.debugMesh[ debugMeshID_cylinder ];
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, assetID_texture_whiteTexture ).bufferID;
	addRenderObject( renderSystem, meshDrawType_lines, mesh.vertex_index, mesh.vertex_count, position, Vec3( radius, height, radius ), Quat( Vec3( 1.0f, 0.0f, 0.0f ), 0.0f ), texture_bufferID, color );
}

internal void
debug_cylinder( cylinder cylinder, vec4 color ) {
	debug_cylinder( cylinder.origin, cylinder.vector.y, cylinder.radius, color );
}

internal void
DEBUG_genMesh_torus( VERTEX_LIST * vertexList ) {
	int32 vertex_count = 1024;
	int32 vertex_index = vertexList->vertex_count;
	
	flo32 major_radius = 0.2286f;
	flo32 minor_radius = 0.005f;
	
	flo32 sin0[ 3 ] = { sinf( PI * 0.125f ), sinf( PI * 0.25f ), sinf( PI * 0.375f ) };
	flo32 cos0[ 3 ] = { cosf( PI * 0.125f ), cosf( PI * 0.25f ), cosf( PI * 0.375f ) };
	flo32 sin_lookup[] = { 0.0f, sin0[ 0 ], sin0[ 1 ], sin0[ 2 ], 1.0f,  sin0[ 2 ],  sin0[ 1 ],  sin0[ 0 ],  0.0f, -sin0[ 0 ], -sin0[ 1 ], -sin0[ 2 ], -1.0f, -sin0[ 2 ], -sin0[ 1 ], -sin0[ 0 ], 0.0f };
	flo32 cos_lookup[] = { 1.0f, cos0[ 0 ], cos0[ 1 ], cos0[ 2 ], 0.0f, -cos0[ 2 ], -cos0[ 1 ], -cos0[ 0 ], -1.0f, -cos0[ 0 ], -cos0[ 1 ], -cos0[ 2 ],  0.0f,  cos0[ 2 ],  cos0[ 1 ],  cos0[ 0 ], 1.0f };
	
	for( int32 iterA = 0; iterA < 16; iterA++ ) {
		vec2 base_vector  = Vec2( cos_lookup[ iterA ], -sin_lookup[ iterA ] );
		vec3 major_vector = Vec3( base_vector.x, 0.0f, base_vector.y ) * major_radius;
		vec3 minor_vector[ 17 ] = {};
		for( int32 iter = 0; iter < 17; iter++ ) {
			minor_vector[ iter ] = Vec3( base_vector.x * cos_lookup[ iter ], sin_lookup[ iter ], base_vector.y * cos_lookup[ iter ] ) * minor_radius;
		}
		
		for( int32 iter = 0; iter < 16; iter++ ) {
			addVertex( vertexList, major_vector + minor_vector[ iter     ], Vec2( 0.0f, 0.0f ) );
			addVertex( vertexList, major_vector + minor_vector[ iter + 1 ], Vec2( 0.0f, 0.0f ) );
		}
	}
	
	VERTEX_TEXMESH_3D * baseVertex = vertexList->vertex + vertex_index;
	for( int32 iter = 0; iter < 512; iter += 2 ) {
		VERTEX_TEXMESH_3D vertexA = baseVertex[ iter      ];
		VERTEX_TEXMESH_3D vertexB = baseVertex[ iter + 32 ];
		addVertex( vertexList, vertexA );
		addVertex( vertexList, vertexB );		
	}
	
	globalVar_debugSystem.debugMesh[ debugMeshID_torus ] = DebugMesh( meshDrawType_lines, vertex_index, vertex_count );
}

internal void
debug_torus( vec3 position, vec4 color ) {
	RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
	
	DEBUG_MESH mesh = globalVar_debugSystem.debugMesh[ debugMeshID_torus ];
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, assetID_texture_whiteTexture ).bufferID;
	addRenderObject( renderSystem, meshDrawType_lines, mesh.vertex_index, mesh.vertex_count, position, Vec3( 1.0f, 1.0f, 1.0f ), Quat( Vec3( 1.0f, 0.0f, 0.0f ), 0.0f ), texture_bufferID, color );
}

internal void
debug_line( vec3 pointA, vec3 pointB, vec4 color ) {
	RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
	
	if( renderSystem->object_count < RENDER_SYSTEM__MAX_OBJECT_COUNT ) {
		VERTEX_BUFFER * vertexBuffer = renderSystem->vertexBuffer;
		
		uint32 texture_bufferID = getTexture( renderSystem->assetSystem, assetID_texture_whiteTexture ).bufferID;
		uint32 vertex_index     = getTempIndex( vertexBuffer );
		
		flo32 sin_lookup[ 5 ] = { 0.0f, sinf( PI * 0.125f ), sinf( PI * 0.25f ), sinf( PI * 0.375f ), 1.0f };
		flo32 cos_lookup[ 5 ] = { 1.0f, cosf( PI * 0.125f ), cosf( PI * 0.25f ), cosf( PI * 0.375f ), 0.0f };
		
		addVertex( vertexBuffer, pointA, Vec2( 0.0f, 0.0f ) );
		addVertex( vertexBuffer, pointB, Vec2( 0.0f, 0.0f ) );
		
		renderSystem->object[ renderSystem->object_count++ ] = RenderObject3D( meshDrawType_lines, vertex_index, 2, Vec3( 0.0f, 0.0f, 0.0f ), Vec3( 1.0f, 1.0f, 1.0f ), Quat( Vec3( 1.0f, 0.0f, 0.0f ), 0.0f ), texture_bufferID, color );
	} else {
		CONSOLE_STRING( "ERROR. RenderObject list is full. Unable to add RenderObject to RenderSystem." );
	}
}

internal void
debug_lineLoop( vec2 * vertex, int32 vertex_count, vec4 color ) {
	RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
	VERTEX_BUFFER    * vertexBuffer = renderSystem->vertexBuffer;
	uint32 texture_bufferID         = getTexture( renderSystem->assetSystem, assetID_texture_whiteTexture ).bufferID;
	
	int32 vertex_index = getTempIndex( vertexBuffer );
	for( int32 iter = 0; iter < vertex_count; iter++ ) {
		vec2 pos = vertex[ iter ];
		addVertex( vertexBuffer, Vec3( pos.x, 0.0f, pos.y ), Vec2( 0.0f, 0.0f ) );
	}
	
	addRenderObject( renderSystem, meshDrawType_lineLoop, vertex_index, vertex_count, Vec3( 0.0f, 0.0f, 0.0f ), Vec3( 1.0f, 1.0f, 1.0f ), Quat( Vec3( 1.0f, 0.0f, 0.0f ), 0.0f ), texture_bufferID, color );
}

internal void
debug_quad( vec3 * vertex, vec4 color ) {
	RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
	VERTEX_BUFFER    * vertexBuffer = renderSystem->vertexBuffer;
	uint32 texture_bufferID         = getTexture( renderSystem->assetSystem, assetID_texture_whiteTexture ).bufferID;
	
	int32 vertex_index = getTempIndex( vertexBuffer );
	for( int32 iter = 0; iter < 4; iter++ ) {
		addVertex( vertexBuffer, vertex[ iter ], Vec2( 0.0f, 0.0f ) );
	}
	
	addRenderObject( renderSystem, meshDrawType_lineLoop, vertex_index, 4, Vec3( 0.0f, 0.0f, 0.0f ), Vec3( 1.0f, 1.0f, 1.0f ), QuatDef(), texture_bufferID, color );
}

internal void
debug_tri( vec3 * vertex, vec4 color ) {
	RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
	VERTEX_BUFFER    * vertexBuffer = renderSystem->vertexBuffer;
	uint32 texture_bufferID         = getTexture( renderSystem->assetSystem, assetID_texture_whiteTexture ).bufferID;
	
	int32 vertex_index = getTempIndex( vertexBuffer );
	for( int32 iter = 0; iter < 3; iter++ ) {
		addVertex( vertexBuffer, vertex[ iter ], Vec2( 0.0f, 0.0f ) );
	}
	
	addRenderObject( renderSystem, meshDrawType_lineLoop, vertex_index, 3, Vec3( 0.0f, 0.0f, 0.0f ), Vec3( 1.0f, 1.0f, 1.0f ), QuatDef(), texture_bufferID, color );
}



internal void
debug_line( vec2 pointA, vec2 pointB, vec4 color ) {
	debug_line( Vec3( pointA.x, 0.0f, pointA.y ), Vec3( pointB.x, 0.0f, pointB.y ), color );
}

internal void serialize_int32 ( char * buffer, const char * name, int32  value ) { sprintf( buffer, "%s: %d",     name, value ); }
internal void serialize_uint64( char * buffer, const char * name, uint64 value ) { sprintf( buffer, "%s: %llu",   name, value ); }
internal void serialize_boo32 ( char * buffer, const char * name, boo32  value ) { sprintf( buffer, "%s: %d",     name, value ); }
internal void serialize_flo32 ( char * buffer, const char * name, flo32  value ) { sprintf( buffer, "%s: %f",     name, value ); }
internal void serialize_vec2  ( char * buffer, const char * name, vec2   value ) { sprintf( buffer, "%s: %f, %f", name, value.x, value.y ); }
internal void serialize_vec3  ( char * buffer, const char * name, vec3   value ) { sprintf( buffer, "%s: %f, %f, %f", name, value.x, value.y, value.z ); }
internal void serialize_rect  ( char * buffer, const char * name, rect   value ) { sprintf( buffer, "%s: %f, %f, %f, %f", name, value.left, value.bottom, value.right, value.top ); }
internal void serialize_MEMORY ( char * buffer, const char * name, MEMORY memory ) { sprintf( buffer, "%s: base: %llx, size: %llu, used: %llu", name, ( uint64 )memory.base, memory.size, memory.used ); }
internal void serialize_RENDER_OBJECT_2D ( char * buffer, const char * name, RENDER_OBJECT_2D object ) { sprintf( buffer, "%s: vertex_index = %u, texture_bufferID = %u, color: { %.02f, %.02f, %.02f, %.02f }", name, object.vertex_index, object.texture_bufferID, object.color.r, object.color.g, object.color.b, object.color.a ); }

#endif // DEBUG_BUILD
#endif	// STD_INCLUDE_FUNC