
#ifdef	STD_INCLUDE_DECL

#endif	// STD_INCLUDE_DECL
#ifdef	STD_INCLUDE_FUNC

internal void
addTask( RENDER_SYSTEM_3D * renderSystem, RENDER_OBJECT_3D renderObject ) {
	Assert( renderSystem->object_count < RENDER_SYSTEM__MAX_OBJECT_COUNT );
	renderSystem->object[ renderSystem->object_count++ ] = renderObject;
}

internal void
addTask( RENDER_SYSTEM_3D * renderSystem, MESH_DRAW_TYPE drawType, int32 vertex_index, int32 vertex_count, vec3 position, vec3 scale, quat orientation, uint32 texture_bufferID, vec4 color ) {
	RENDER_OBJECT_3D renderObject = { drawType, vertex_index, vertex_count, position, scale, orientation, texture_bufferID, color };
	addTask( renderSystem, renderObject );
}

internal void
addTask( RENDER_SYSTEM_3D * renderSystem, RENDER_MESH mesh, uint32 texture_bufferID, vec3 position, vec3 scale, quat orientation, vec4 color ) {
	VERTEX_BUFFER * vertexBuffer = renderSystem->vertexBuffer;
	
	uint32 vertex_index = getTempIndex( vertexBuffer );
	addTemp( vertexBuffer, mesh.vertex, mesh.vertex_count );
	addTask( renderSystem, mesh.drawType, vertex_index, mesh.vertex_count, position, scale, orientation, texture_bufferID, color );
}

internal void
addTask( RENDER_SYSTEM_3D * renderSystem, PERM_MESH mesh, uint32 texture_bufferID, vec3 position, vec3 scale, quat orientation, vec4 color ) {
	addTask( renderSystem, mesh.drawType, mesh.vertex_index, mesh.vertex_count, position, scale, orientation, texture_bufferID, color );
}

internal void
addTask( RENDER_SYSTEM_2D * renderSystem, MESH_DRAW_TYPE drawType, uint32 texture_bufferID, rect bound, vec4 color ) {
	beginTask( renderSystem, drawType, texture_bufferID, color );
	addQuad( renderSystem, bound, Rect( 0.0f, 0.0f, 1.0f, 1.0f ) );
	endTask( renderSystem );
}

internal void
drawRect( RENDER_SYSTEM_2D * renderSystem, ASSET_ID textureID, rect bound, vec4 color ) {
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, textureID ).bufferID;
	addTask( renderSystem, meshDrawType_triangleFan, texture_bufferID, bound, color );
}

internal void
drawMesh( RENDER_SYSTEM_3D * renderSystem, ASSET_ID meshID, ASSET_ID textureID, vec3 position, vec3 scale, quat orientation, vec4 color ) {
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, textureID ).bufferID;
	RENDER_MESH        mesh = getMesh(    renderSystem->assetSystem,    meshID );
	addTask( renderSystem, mesh, texture_bufferID, position, scale, orientation, color );
}

internal void
drawMesh( RENDER_SYSTEM_3D * renderSystem, RENDER_MESH mesh, ASSET_ID textureID, vec3 position, vec3 scale, quat orientation, vec4 color ) {	
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, textureID ).bufferID;
	addTask( renderSystem, mesh, texture_bufferID, position, scale, orientation, color );
}

internal void
drawMesh( RENDER_SYSTEM_3D * renderSystem, ASSET_ID meshID, ASSET_ID textureID ) {
	drawMesh( renderSystem, meshID, textureID, Vec3( 0.0f, 0.0f, 0.0f ), Vec3( 1.0f, 1.0f, 1.0f ), QuatDef(), COLOR_WHITE );
}

internal void
drawPermMesh( RENDER_SYSTEM_3D * renderSystem, PERM_MESH mesh, ASSET_ID textureID, vec3 position, vec3 scale, quat orientation, vec4 color ) {
	uint32 texture_bufferID = getTexture( renderSystem->assetSystem, textureID ).bufferID;
	if( renderSystem->object_count < RENDER_SYSTEM__MAX_OBJECT_COUNT ) {
		renderSystem->object[ renderSystem->object_count++ ] = RenderObject3D( mesh.drawType, mesh.vertex_index, mesh.vertex_count, position, scale, orientation, texture_bufferID, color );
	} else {
		CONSOLE_STRING( "ERROR. RenderObject list is full. Unable to add RenderObject to RenderSystem." );
	}
}

#endif	// STD_INCLUDE_FUNC
