
#ifdef	STD_INCLUDE_DECL

typedef boo32 ( READ_FILE_CHUNK  )( uint64 fileHandle, uint32 offset, uint32 size, void * buffer );
typedef boo32 ( READ_ASSET_CHUNK )( uint64 fileHandle, uint32 offset, uint32 size, void * buffer );

struct PLATFORM {
	MOUSE_STATE mouse;
	KEYBOARD_STATE keyboard;
	
	MEMORY permMemory;
	MEMORY tempMemory;
	
	READ_FILE_CHUNK  * readFile_chunk;
	READ_ASSET_CHUNK * readAsset_chunk;
	
	ASSET_SYSTEM   assetSystem;
	RENDER_SYSTEM_2D renderSystem2D;
	RENDER_SYSTEM_3D renderSystem3D;
	
	VERTEX_BUFFER vertexBuffer;
	
	flo32 targetSec;
};

#endif	// STD_INCLUDE_DECL
#ifdef	STD_INCLUDE_FUNC

internal boo32
win32_readFile_chunk( uint64 fileHandle, uint32 offset, uint32 size, void * buffer ) {
	boo32 result = false;
	if( fileHandle ) {
		// TODO: check that file handle is still open
		DWORD error = SetFilePointer( ( HANDLE )fileHandle, offset, 0, FILE_BEGIN );
		if( error != INVALID_SET_FILE_POINTER ) {
			DWORD bytesRead = 0;
			ReadFile( ( HANDLE )fileHandle, buffer, size, &bytesRead, 0 );
			if( bytesRead == size ) {
				result = true;
			} else {
				CONSOLE_STRING( "ERROR. Error reading in requested file chunk." );
			}
		} else {
			CONSOLE_STRING( "ERROR. Unable to set file pointer in requested file." );
		}
	} else {
		CONSOLE_STRING( "ERROR. fileHandle is not valid." );
	}
	return result;
}

internal int64
win32_getPerfCount() {
	LARGE_INTEGER perfCounter = {};
	QueryPerformanceCounter( &perfCounter );
	
	int64 result = *( ( int64 * )&perfCounter );
	return result;
}

internal flo32
win32_getMSElapsed( int64 startCounter, int64 endCounter, int64 perfFrequency ) {
	LARGE_INTEGER counterA  = *( ( LARGE_INTEGER * )&startCounter  );
	LARGE_INTEGER counterB  = *( ( LARGE_INTEGER * )&endCounter    );
	LARGE_INTEGER frequency = *( ( LARGE_INTEGER * )&perfFrequency );
	flo32 result = 1000.0f * ( ( flo32 )( counterB.QuadPart - counterA.QuadPart ) / ( flo32 )( frequency.QuadPart ) );
	return result;
}

LRESULT CALLBACK
win32_WindowProc( HWND window, uint32 message, WPARAM wParam, LPARAM lParam ) {
	LRESULT result = 0;
	switch( message ) {
		case WM_ACTIVATEAPP: {
		} break;

		case WM_CLOSE: {
		} break;

		case WM_DESTROY: {
		} break;
		
		case WM_SYSKEYDOWN:
		case WM_SYSKEYUP:
		case WM_KEYDOWN:
		case WM_KEYUP: {
			Assert( !"ERROR! Keyboard message was passed directly to WindowProc! All keyboard messages should be handled directly by the game!" );
		} break;
		
		default: {
			result = DefWindowProc( window, message, wParam, lParam );
		} break;
	}
	return result;
}

internal void
win32_processPendingMessages( PLATFORM * platform, HWND windowHandle ) {
	// PROFILE_FUNCTION();
	
	KEYBOARD_STATE * keyboard = &platform->keyboard;
	keyboard->event_count = 0;
	
	MOUSE_STATE * mouse = &platform->mouse;
	MOUSE_endOfFrame( mouse );
	
	MSG message;
	while( PeekMessage( &message, 0, 0, 0, PM_REMOVE ) ) {
		switch( message.message ) {
			case WM_MOUSEWHEEL: {
				int16 wParam_hi = ( ( message.wParam >> 16 ) & 0xFFFF );
				int32 wheelClicks = wParam_hi / 120;
				mouse->wheelClicks = wheelClicks;
			} break;
			
			case WM_LBUTTONDOWN:
			case WM_LBUTTONUP: {
				platform->mouse.control_isDown = ( message.wParam & MK_CONTROL );
				platform->mouse.shift_isDown = ( message.wParam & MK_SHIFT );
				processButtonEvent( mouse, mouseButton_left, ( message.wParam & MK_LBUTTON ) );
			} break;
			
			case WM_MBUTTONDOWN:
			case WM_MBUTTONUP: {
				platform->mouse.control_isDown = ( message.wParam & MK_CONTROL );
				platform->mouse.shift_isDown = ( message.wParam & MK_SHIFT );
				processButtonEvent( mouse, mouseButton_middle, ( message.wParam & MK_MBUTTON ) );
			} break;
			
			case WM_RBUTTONUP:
			case WM_RBUTTONDOWN: {
				platform->mouse.control_isDown = ( message.wParam & MK_CONTROL );
				platform->mouse.shift_isDown = ( message.wParam & MK_SHIFT );
				processButtonEvent( mouse, mouseButton_right, ( message.wParam & MK_RBUTTON ) );
			} break;
			
			case WM_SYSKEYDOWN:
			case WM_SYSKEYUP:
			case WM_KEYDOWN:
			case WM_KEYUP: {
				boo32 wasDown = ( ( message.lParam & ( 1 << 30 ) ) != 0 );
				boo32 isDown = ( ( message.lParam & ( 1 << 31 ) ) == 0 );
				if( isDown != wasDown ) {
					switch( message.wParam ) {
						case VK_BACK: 		{ addKeyEvent( keyboard, keyCode_backspace, 	isDown ); } break;
						case VK_TAB: 		{ addKeyEvent( keyboard, keyCode_tab, 			isDown ); } break;
						case VK_RETURN:	{ addKeyEvent( keyboard, keyCode_enter, 		isDown ); } break;
						case VK_SHIFT: 	{
							keyboard->shiftIsDown = isDown;
							addKeyEvent( keyboard, keyCode_shift, isDown );
						} break;
						case VK_CONTROL: 	{
							keyboard->controlIsDown = isDown;
							addKeyEvent( keyboard, keyCode_control, isDown );
						} break;
						case VK_MENU: 		{ addKeyEvent( keyboard, keyCode_alt, 			isDown ); } break;
						case VK_CAPITAL: 	{ addKeyEvent( keyboard, keyCode_capsLock, 	isDown ); } break;
						case VK_ESCAPE: 	{ addKeyEvent( keyboard, keyCode_escape, 		isDown ); } break;
						case VK_SPACE: 	{ addKeyEvent( keyboard, keyCode_space, 		isDown ); } break;
						case VK_PRIOR: 	{ addKeyEvent( keyboard, keyCode_pageUp, 		isDown ); } break;
						case VK_NEXT: 		{ addKeyEvent( keyboard, keyCode_pageDown, 	isDown ); } break;
						case VK_END: 		{ addKeyEvent( keyboard, keyCode_end, 			isDown ); } break;
						case VK_HOME: 		{ addKeyEvent( keyboard, keyCode_home, 			isDown ); } break;
						case VK_LEFT: 		{ addKeyEvent( keyboard, keyCode_left, 			isDown ); } break;
						case VK_RIGHT: 	{ addKeyEvent( keyboard, keyCode_right, 		isDown ); } break;
						case VK_UP: 		{ addKeyEvent( keyboard, keyCode_up, 			isDown ); } break;
						case VK_DOWN: 		{ addKeyEvent( keyboard, keyCode_down, 			isDown ); } break;
						case VK_INSERT: 	{ addKeyEvent( keyboard, keyCode_insert, 		isDown ); } break;
						case VK_DELETE: 	{ addKeyEvent( keyboard, keyCode_delete, 		isDown ); } break;
						case 0x30: { addKeyEvent( keyboard, keyCode_0, isDown, keyboard->shiftIsDown ); } break;
						case 0x31: { addKeyEvent( keyboard, keyCode_1, isDown, keyboard->shiftIsDown ); } break;
						case 0x32: { addKeyEvent( keyboard, keyCode_2, isDown, keyboard->shiftIsDown ); } break;
						case 0x33: { addKeyEvent( keyboard, keyCode_3, isDown, keyboard->shiftIsDown ); } break;
						case 0x34: { addKeyEvent( keyboard, keyCode_4, isDown, keyboard->shiftIsDown ); } break;
						case 0x35: { addKeyEvent( keyboard, keyCode_5, isDown, keyboard->shiftIsDown ); } break;
						case 0x36: { addKeyEvent( keyboard, keyCode_6, isDown, keyboard->shiftIsDown ); } break;
						case 0x37: { addKeyEvent( keyboard, keyCode_7, isDown, keyboard->shiftIsDown ); } break;
						case 0x38: { addKeyEvent( keyboard, keyCode_8, isDown, keyboard->shiftIsDown ); } break;
						case 0x39: { addKeyEvent( keyboard, keyCode_9, isDown, keyboard->shiftIsDown ); } break;
						case 0x41: { addKeyEvent( keyboard, keyCode_a, isDown, keyboard->shiftIsDown ); } break;
						case 0x42: { addKeyEvent( keyboard, keyCode_b, isDown, keyboard->shiftIsDown ); } break;
						case 0x43: { addKeyEvent( keyboard, keyCode_c, isDown, keyboard->shiftIsDown ); } break;
						case 0x44: { addKeyEvent( keyboard, keyCode_d, isDown, keyboard->shiftIsDown ); } break;
						case 0x45: { addKeyEvent( keyboard, keyCode_e, isDown, keyboard->shiftIsDown ); } break;
						case 0x46: { addKeyEvent( keyboard, keyCode_f, isDown, keyboard->shiftIsDown ); } break;
						case 0x47: { addKeyEvent( keyboard, keyCode_g, isDown, keyboard->shiftIsDown ); } break;
						case 0x48: { addKeyEvent( keyboard, keyCode_h, isDown, keyboard->shiftIsDown ); } break;
						case 0x49: { addKeyEvent( keyboard, keyCode_i, isDown, keyboard->shiftIsDown ); } break;
						case 0x4A: { addKeyEvent( keyboard, keyCode_j, isDown, keyboard->shiftIsDown ); } break;
						case 0x4B: { addKeyEvent( keyboard, keyCode_k, isDown, keyboard->shiftIsDown ); } break;
						case 0x4C: { addKeyEvent( keyboard, keyCode_l, isDown, keyboard->shiftIsDown ); } break;
						case 0x4D: { addKeyEvent( keyboard, keyCode_m, isDown, keyboard->shiftIsDown ); } break;
						case 0x4E: { addKeyEvent( keyboard, keyCode_n, isDown, keyboard->shiftIsDown ); } break;
						case 0x4F: { addKeyEvent( keyboard, keyCode_o, isDown, keyboard->shiftIsDown ); } break;
						case 0x50: { addKeyEvent( keyboard, keyCode_p, isDown, keyboard->shiftIsDown ); } break;
						case 0x51: { addKeyEvent( keyboard, keyCode_q, isDown, keyboard->shiftIsDown ); } break;
						case 0x52: { addKeyEvent( keyboard, keyCode_r, isDown, keyboard->shiftIsDown ); } break;
						case 0x53: { addKeyEvent( keyboard, keyCode_s, isDown, keyboard->shiftIsDown ); } break;
						case 0x54: { addKeyEvent( keyboard, keyCode_t, isDown, keyboard->shiftIsDown ); } break;
						case 0x55: { addKeyEvent( keyboard, keyCode_u, isDown, keyboard->shiftIsDown ); } break;
						case 0x56: { addKeyEvent( keyboard, keyCode_v, isDown, keyboard->shiftIsDown ); } break;
						case 0x57: { addKeyEvent( keyboard, keyCode_w, isDown, keyboard->shiftIsDown ); } break;
						case 0x58: { addKeyEvent( keyboard, keyCode_x, isDown, keyboard->shiftIsDown ); } break;
						case 0x59: { addKeyEvent( keyboard, keyCode_y, isDown, keyboard->shiftIsDown ); } break;
						case 0x5A: { addKeyEvent( keyboard, keyCode_z, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_1: { addKeyEvent( keyboard, keyCode_semicolon, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_PLUS: { addKeyEvent( keyboard, keyCode_equal, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_COMMA: { addKeyEvent( keyboard, keyCode_comma, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_MINUS: { addKeyEvent( keyboard, keyCode_dash, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_PERIOD: { addKeyEvent( keyboard, keyCode_period, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_2: { addKeyEvent( keyboard, keyCode_forwardSlash, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_3: { addKeyEvent( keyboard, keyCode_tilde, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_4: { addKeyEvent( keyboard, keyCode_openBracket, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_5: { addKeyEvent( keyboard, keyCode_backslash, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_6: { addKeyEvent( keyboard, keyCode_closeBracket, isDown, keyboard->shiftIsDown ); } break;
						case VK_OEM_7: { addKeyEvent( keyboard, keyCode_quote, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD0: { addKeyEvent( keyboard, keyCode_num0, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD1: { addKeyEvent( keyboard, keyCode_num1, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD2: { addKeyEvent( keyboard, keyCode_num2, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD3: { addKeyEvent( keyboard, keyCode_num3, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD4: { addKeyEvent( keyboard, keyCode_num4, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD5: { addKeyEvent( keyboard, keyCode_num5, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD6: { addKeyEvent( keyboard, keyCode_num6, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD7: { addKeyEvent( keyboard, keyCode_num7, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD8: { addKeyEvent( keyboard, keyCode_num8, isDown, keyboard->shiftIsDown ); } break;
						case VK_NUMPAD9: { addKeyEvent( keyboard, keyCode_num9, isDown, keyboard->shiftIsDown ); } break;
						default: {
							char string[ 128 ] = {};
							sprintf( string, "Key message received, but not processed: %lu %s\n", message.wParam, ( isDown ? "PRESSED" : "RELEASED" ) );
							OutputDebugString( string );
						} break;
					};
				}
			} break;
			
			default: {
				TranslateMessage( &message );
				DispatchMessage( &message );
			} break;
		}
	}
}

internal void
win32_endApplication( void * platform, void * appState ) {
	APP_STATE * myAppState = ( APP_STATE * )appState;
	myAppState->isRunning = false;
}

internal void
render2D( RENDER_SYSTEM_2D * renderSystem, int32 shader_uniform_modColor ) {
	PROFILE_FUNCTION();
	
	VERTEX_BUFFER * vertexBuffer = renderSystem->vertexBuffer;
	
	// glBufferSubData( GL_ARRAY_BUFFER, sizeof( VERTEX_TEXMESH_3D ) * renderSystem->base_index, sizeof( VERTEX_TEXMESH_3D ) * vertexList->vertex_count, vertexList->vertex );

	for( int32 object_index = 0; object_index < renderSystem->object_count; object_index++ ) {
		RENDER_OBJECT_2D object = renderSystem->object[ object_index ];
		
		GLuint drawType = 0;
		switch( object.drawType ) {
			case meshDrawType_triangles:   { drawType = GL_TRIANGLES;    } break;
			case meshDrawType_triangleFan: { drawType = GL_TRIANGLE_FAN; } break;
			case meshDrawType_lineLoop:    { drawType = GL_LINE_LOOP;    } break;
			default: { InvalidCodePath; } break;
		}
		
		glBindTexture( GL_TEXTURE_2D, object.texture_bufferID );
		glUniform4fv( shader_uniform_modColor, 1, object.color.elem );
		glDrawArrays( drawType, object.vertex_index, object.vertex_count );
	}
	
	renderSystem->object_count = 0;
}

internal void
render3D( RENDER_SYSTEM_3D * renderSystem, int32 shader_uniform_transform, int32 shader_uniform_modColor ) {
	PROFILE_FUNCTION();
	
	VERTEX_BUFFER * vertexBuffer = renderSystem->vertexBuffer;
	
	for( int32 object_index = 0; object_index < renderSystem->object_count; object_index++ ) {
		RENDER_OBJECT_3D object = renderSystem->object[ object_index ];
		
		GLuint drawType = 0;
		switch( object.drawType ) {
			case meshDrawType_triangles:   { drawType = GL_TRIANGLES;    } break;
			case meshDrawType_triangleFan: { drawType = GL_TRIANGLE_FAN; } break;
			case meshDrawType_lines:       { drawType = GL_LINES;        } break;
			case meshDrawType_lineLoop:    { drawType = GL_LINE_LOOP;    } break;
			default: { InvalidCodePath; } break;
		}
		
		mat4 transform = renderSystem->camera_transform * genModelTransform( object.position, object.scale, object.orientation );
		
		glBindTexture( GL_TEXTURE_2D, object.texture_bufferID );
		glUniform4fv( shader_uniform_modColor, 1, object.color.elem );
		glUniformMatrix4fv( shader_uniform_transform, 1, GL_FALSE, transform.elem );
		glDrawArrays( drawType, object.vertex_index, object.vertex_count );
	}
	
	renderSystem->object_count = 0;
}

typedef HGLRC WINAPI wgl_create_context_attribs_arb( HDC hDC, HGLRC hShareContext, const int * attribList );
typedef BOOL  WINAPI wgl_choose_pixel_format_arb( HDC hdc, const int * piAttribIList, const FLOAT * pfAttribFList, UINT nMaxFormats, int * piFormats, UINT * nNumFormats );

int32 CALLBACK
WinMain( HINSTANCE instance, HINSTANCE prevInstance, LPSTR commandLine, int32 windowShowCode ) {
	srand( ( uint32 )time( 0 ) );
	
	boo32 sleepFreqIsHiRes = ( timeBeginPeriod( 1 ) == TIMERR_NOERROR );
	
	LARGE_INTEGER perfFrequency;
	QueryPerformanceFrequency( &perfFrequency );
	
	WNDCLASS windowClass = {};
	windowClass.lpfnWndProc = win32_WindowProc;
	windowClass.hInstance = instance;
	windowClass.lpszClassName = "WindowClass";
	
	if( RegisterClass( &windowClass ) ) {
		int32 screen_width = GetSystemMetrics( SM_CXSCREEN );
		int32 screen_height = GetSystemMetrics( SM_CYSCREEN );
		
		// HWND window = CreateWindowEx( 0, "WindowClass", "Caption", WS_VISIBLE, 0, 0, screen_width, screen_height, 0, 0, instance, 0 ); // WINDOWED
		HWND window = CreateWindowEx( 0, "WindowClass", "Caption", WS_EX_TOPMOST | WS_VISIBLE | WS_POPUP, 0, 0, screen_width, screen_height, 0, 0, instance, 0 ); // FULLSCREEN
		ShowCursor( FALSE );

		if( window ) {
			RECT client_rect = {};
			GetClientRect( window, &client_rect );
			
			int32 client_width = client_rect.right - client_rect.left;
			int32 client_height = client_rect.bottom - client_rect.top;
			
			vec2 client_dim = Vec2( ( flo32 )client_width, ( flo32 )client_height );
			vec2 client_halfDim = client_dim * 0.5f;
			rect client_bound = Rect( ( flo32 )client_rect.left, ( flo32 )client_rect.top, ( flo32 )client_rect.right, ( flo32 )client_rect.bottom );
			
			int32 vertexBuffer_tempCount  = RENDER_SYSTEM__MAX_VERTEX_COUNT * 2;
			int32 vertexBuffer_permCount  = RENDER_SYSTEM__MAX_VERTEX_COUNT;
#if DEBUG_BUILD
			      vertexBuffer_tempCount +=  DEBUG_SYSTEM__MAX_VERTEX_COUNT * 2;
#endif // DEBUG_BUILD
			
			MEMORY platformMemory = {};
			platformMemory.size = ( sizeof( APP_STATE ) + sizeof( PLATFORM ) + APP_permMemorySize + APP_tempMemorySize );
			platformMemory.base = VirtualAlloc( 0, platformMemory.size, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
			
			APP_STATE * appState = _pushType( &platformMemory, APP_STATE );
			appState->isRunning = true;
			
			PLATFORM * win32_platform = _pushType( &platformMemory, PLATFORM );
			win32_platform->permMemory = { _pushSize( &platformMemory, APP_permMemorySize ), APP_permMemorySize };
			win32_platform->tempMemory = { _pushSize( &platformMemory, APP_tempMemorySize ), APP_tempMemorySize };
			win32_platform->readFile_chunk  = win32_readFile_chunk;
			win32_platform->readAsset_chunk = win32_readFile_chunk;
			win32_platform->targetSec = 1.0f / APP_targetFPS;
			
			VERTEX_BUFFER * vertexBuffer = &win32_platform->vertexBuffer;
			vertexBuffer->temp_maxCount  = vertexBuffer_tempCount;
			vertexBuffer->perm_baseIndex = vertexBuffer_tempCount * 3;
			vertexBuffer->perm_maxCount  = vertexBuffer_permCount;
			
			ASSET_SYSTEM     *  assetSystem   = &win32_platform->assetSystem;
			RENDER_SYSTEM_2D * renderSystem2D = &win32_platform->renderSystem2D;
			RENDER_SYSTEM_3D * renderSystem3D = &win32_platform->renderSystem3D;
			renderSystem2D->assetSystem  = assetSystem;
			renderSystem3D->assetSystem  = assetSystem;
			renderSystem2D->vertexBuffer = vertexBuffer;
			renderSystem3D->vertexBuffer = vertexBuffer;
			
			HANDLE assetFile_handle = CreateFile( "assets.pak", GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0 );
			if( assetFile_handle != INVALID_HANDLE_VALUE ) {
				ASSET_SYSTEM * assetSystem = &win32_platform->assetSystem;
				
				assetSystem->platform = win32_platform;
				assetSystem->assetFile_handle = ( uint64 )assetFile_handle;
				
				uint32 read_size = _arrayBytes( assetSystem->entry );
				DWORD bytesRead = 0;
				
				ReadFile( assetFile_handle, assetSystem->entry, read_size, &bytesRead, 0 );
				Assert( bytesRead == read_size );
			} else {
				OutputDebugString( "ERROR: Unable to open asset file: assets.pak" );
			}
			
			HDC win32_deviceContext = GetDC( window );
			
			vec2 app_dim = client_dim;
			vec2 app_halfDim = app_dim * 0.5f;
			rect app_bound = rectBLD( Vec2( 0.0f, 0.0f ), app_dim );
			appState->app_dim = app_dim;
			appState->app_halfDim = app_halfDim;
			appState->app_bound = app_bound;
			
			{
#define WGL_DRAW_TO_WINDOW_ARB                  0x2001
#define WGL_ACCELERATION_ARB                    0x2003
#define WGL_SUPPORT_OPENGL_ARB                  0x2010
#define WGL_DOUBLE_BUFFER_ARB                   0x2011
#define WGL_PIXEL_TYPE_ARB                      0x2013

#define WGL_TYPE_RGBA_ARB                       0x202B
#define WGL_FULL_ACCELERATION_ARB               0x2027

            wgl_choose_pixel_format_arb    * wglChoosePixelFormatARB    = ( wgl_choose_pixel_format_arb    * )wglGetProcAddress( "wglChoosePixelFormatARB"    );
            wgl_create_context_attribs_arb * wglCreateContextAttribsARB = ( wgl_create_context_attribs_arb * )wglGetProcAddress( "wglCreateContextAttribsARB" );
				
				int32 SuggestedPixelFormatIndex = 0;
				GLuint ExtendedPick = 0;
				
				if( wglChoosePixelFormatARB ) {
					int32 IntAttribList[] = {
						WGL_DRAW_TO_WINDOW_ARB, GL_TRUE, // 0
						WGL_ACCELERATION_ARB,   WGL_FULL_ACCELERATION_ARB, // 1
						WGL_SUPPORT_OPENGL_ARB, GL_TRUE, // 2
						WGL_DOUBLE_BUFFER_ARB,  GL_TRUE, // 3
						WGL_PIXEL_TYPE_ARB,     WGL_TYPE_RGBA_ARB, // 4
						0,
					};

					wglChoosePixelFormatARB( win32_deviceContext, IntAttribList, 0, 1, &SuggestedPixelFormatIndex, &ExtendedPick );
				}

				if( !ExtendedPick ) {
					PIXELFORMATDESCRIPTOR DesiredPixelFormat = {};
					DesiredPixelFormat.nSize = sizeof(DesiredPixelFormat);
					DesiredPixelFormat.nVersion = 1;
					DesiredPixelFormat.iPixelType = PFD_TYPE_RGBA;
					DesiredPixelFormat.dwFlags = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER;
					DesiredPixelFormat.cColorBits = 32;
					DesiredPixelFormat.cAlphaBits = 8;
					DesiredPixelFormat.cDepthBits = 32;
					DesiredPixelFormat.iLayerType = PFD_MAIN_PLANE;

					SuggestedPixelFormatIndex = ChoosePixelFormat( win32_deviceContext, &DesiredPixelFormat );
				}

				PIXELFORMATDESCRIPTOR SuggestedPixelFormat;
				DescribePixelFormat( win32_deviceContext, SuggestedPixelFormatIndex, sizeof( SuggestedPixelFormat ), &SuggestedPixelFormat );
				     SetPixelFormat( win32_deviceContext, SuggestedPixelFormatIndex,                                 &SuggestedPixelFormat );
			}
			
			HGLRC win32_renderContext = wglCreateContext( win32_deviceContext );
			boo32 success_setContext = ( boo32 )wglMakeCurrent( win32_deviceContext, win32_renderContext );
			Assert( success_setContext );
			
			GL_win32_getExtensions();
			
			typedef BOOL ( SWAP_INTERVAL )( int32 interval );
			SWAP_INTERVAL * swapInterval = ( SWAP_INTERVAL * )wglGetProcAddress( "wglSwapIntervalEXT" );
			if( swapInterval ) {
				swapInterval( 1 );
			} else {
				OutputDebugString( "VSync not supported.\n" );
			}
			
			glBindFramebuffer( GL_FRAMEBUFFER, 0 );
			glViewport( 0, 0, client_width, client_height );
			glClearColor( 0.8f, 0.8f, 0.8f, 1.0f );
			glClearDepth( 1.0f );
			glBlendFunc( GL_ONE, GL_ONE_MINUS_SRC_ALPHA );
			
			glEnable( GL_CULL_FACE );
			glCullFace( GL_BACK );
			glFrontFace( GL_CCW );
			
			int32 mainShader_transform = -1;
			int32 mainShader_modColor  = -1;
			{ // main shader			
				char * shader_vertexCode = "\
				#version 330 core\n\
				uniform mat4 transform;\n\
				layout( location = 0 ) in vec3 inPosition;\n\
				layout( location = 1 ) in vec2 inTexCoord;\n\
				out vec2 textureCoord;\n\
				void main( void ) {\n\
					gl_Position = transform * vec4( inPosition, 1.0f );\n\
					textureCoord = inTexCoord;\n\
				}\n";
					
				char * shader_fragmentCode = "\
				#version 330 core\n\
				uniform sampler2D textureSampler;\n\
				uniform vec4 modColor;\n\
				in vec2 textureCoord;\n\
				out vec4 outColor;\n\
				void main( void ) {\n\
					outColor = modColor * texture( textureSampler, textureCoord.st );\n\
				}\n";
				
				CREATE_SHADER_RESULT shader = createShader( shader_vertexCode, shader_fragmentCode );
				Assert( shader.isValid );
				
				glUseProgram( shader.program );
				mainShader_modColor  = glGetUniformLocation( shader.program, "modColor" );
				mainShader_transform = glGetUniformLocation( shader.program, "transform" );
				
				glUniform1i( glGetUniformLocation( shader.program, "textureSampler" ), 0 );
			}
			mat4 screen_transform = mat4_orthographic( client_bound.left, client_bound.bottom, client_bound.right, client_bound.top, -1.0f, 1.0f );
			mat4 perspective_transform = mat4_perspective( client_dim, 50.0f, 0.1f, 1000.0f );
			
			{ // gen default texture
				uint8 defaultTexture[] = { 80, 80, 80, 255,   255, 0, 255, 255,   255, 0, 255, 255,   80, 80, 80, 255 };
				int32 width = 2;
				int32 height = 2;
				
				uint32 texture_bufferID = 0;
				glGenTextures( 1, &texture_bufferID );
				glBindTexture( GL_TEXTURE_2D, texture_bufferID );
				glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, defaultTexture );
				glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
				glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
				glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
				glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
				
				assetSystem->defaultTexture = RenderTexture( texture_bufferID, width, height );
			}
			
			{ // gen default font
				genDefaultMonoFont( assetSystem, &win32_platform->tempMemory );
				genDefaultMesh( assetSystem );
			}
			
#if DEBUG_BUILD
			int32 	 debugSystem_sizeInBytes = DEBUG_SYSTEM__SIZE_IN_BYTES;
			MEMORY	_debugSystem_memory = Memory( VirtualAlloc( 0, debugSystem_sizeInBytes, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE ), debugSystem_sizeInBytes );
			MEMORY *  debugSystem_memory = &_debugSystem_memory;
			globalVar_debugSystem.renderSystem2D 					=  _pushType( debugSystem_memory, RENDER_SYSTEM_2D );
			globalVar_debugSystem.renderSystem3D               =  _pushType( debugSystem_memory, RENDER_SYSTEM_3D );
			globalVar_advanceHeight 					= assetSystem->defaultMonoFont.font.advanceHeight;
			globalVar_debugSystem.renderSystem2D->assetSystem  = assetSystem;
			globalVar_debugSystem.renderSystem3D->assetSystem  = assetSystem;
			globalVar_debugSystem.renderSystem2D->vertexBuffer = vertexBuffer;
			globalVar_debugSystem.renderSystem3D->vertexBuffer = vertexBuffer;
			
			int32 	 profileSystem_sizeInBytes = PROFILE_SYSTEM__SIZE_IN_BYTES;
			MEMORY	_profileSystem_memory = Memory( VirtualAlloc( 0, profileSystem_sizeInBytes, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE ), profileSystem_sizeInBytes );
			MEMORY *	 profileSystem_memory = &_profileSystem_memory;
			globalVar_advanceHeight 	= assetSystem->defaultMonoFont.font.advanceHeight;
			globalVar_profileSystem.getMSElapsed = win32_getMSElapsed;
			globalVar_profileSystem.eventDatabase.eventList 		= _pushArray( profileSystem_memory, PROFILE_EVENT_LIST, PROFILE_EVENT_DATABASE_MAX_COUNT );
			globalVar_profileSystem.eventDatabase.maxCount 			= PROFILE_EVENT_DATABASE_MAX_COUNT;
			globalVar_profileSystem.eventDatabase.selectIndex 		= -1;
			globalVar_profileSystem.eventList_currentFrame			= _pushType( profileSystem_memory, PROFILE_EVENT_LIST );
			globalVar_profileSystem.nodeList_currentFrame			= _pushType( profileSystem_memory, PROFILE_NODE_LIST );
			globalVar_profileSystem.eventList_currentFrame->frameClockStart = __rdtsc();
			
			globalVar_profileSystem.eventDatabase_clocks_currentFrame.frameStart 	= __rdtsc();
			globalVar_profileSystem.eventDatabase_clocks_currentFrame.platformStart	= __rdtsc();
			globalVar_profileSystem.eventDatabase_clocks_currentFrame.appStart 		= __rdtsc();
			globalVar_profileSystem.eventDatabase_clocks_currentFrame.renderStart 	= __rdtsc();
			globalVar_profileSystem.eventDatabase_clocks_currentFrame.frameEnd 		= __rdtsc();
			globalVar_profileSystem.eventDatabase_clocks_currentFrame.counter_frameStart = win32_getPerfCount();
			globalVar_profileSystem.eventDatabase_clocks_currentFrame.counter_frameEnd   = win32_getPerfCount();
			
			globalVar_profileSystem.eventDatabase_clocks = _pushArray( profileSystem_memory, PROFILE_EVENT_DATABASE_CLOCKS, PROFILE_EVENT_DATABASE_MAX_COUNT );
			
			globalVar_profileSystem.eventDatabase_posOffset 				= 4.0f;
			globalVar_profileSystem.eventDatabase_width 						= client_width - 2.0f * globalVar_profileSystem.eventDatabase_posOffset;
			globalVar_profileSystem.eventDatabase_height 					= 300.0f;
			globalVar_profileSystem.eventDatabase_scrollActive 			= false;
			globalVar_profileSystem.frameBreakdown_scrollActive_bar 		= false;
			globalVar_profileSystem.frameBreakdown_scrollActive_left 	= false;
			globalVar_profileSystem.frameBreakdown_scrollActive_right 	= false;
			globalVar_profileSystem.frameBreakdown_scrollX_left  			= globalVar_profileSystem.eventDatabase_posOffset;
			globalVar_profileSystem.frameBreakdown_scrollX_right 			= client_width - globalVar_profileSystem.eventDatabase_posOffset;
#endif
			uint32    mapFlags = GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT | GL_MAP_COHERENT_BIT;
			uint32 createFlags = mapFlags | GL_DYNAMIC_STORAGE_BIT;
			
			uint32 vertexBuffer_id   = 0;
			uint32 vertexBuffer_size = ( vertexBuffer_tempCount * 3 + vertexBuffer_permCount ) * sizeof( VERTEX_TEXMESH_3D );
			glGenBuffers( 1, &vertexBuffer_id );
			glBindBuffer( GL_ARRAY_BUFFER, vertexBuffer_id );
			glBufferStorage( GL_ARRAY_BUFFER, vertexBuffer_size, 0, createFlags );
			vertexBuffer->buffer_base = ( VERTEX_TEXMESH_3D * )glMapBufferRange( GL_ARRAY_BUFFER, 0, vertexBuffer_size, mapFlags );
			
			GLsync openGLSync[ 3 ] = {};
			for( int32 iter = 0; iter < 3; iter++ ) {
				openGLSync[ iter ] = glFenceSync( GL_SYNC_GPU_COMMANDS_COMPLETE, 0 );
			}

			uint32 stride = sizeof( VERTEX_TEXMESH_3D );
			glVertexAttribPointer( 0, 3, GL_FLOAT, GL_FALSE, stride, ( GLvoid * )( 0 ) );
			glEnableVertexAttribArray( 0 );
			
			glVertexAttribPointer( 1, 2, GL_FLOAT, GL_FALSE, stride, ( GLvoid * )( sizeof( flo32 ) * 3 ) );
			glEnableVertexAttribArray( 1 );

#if DEBUG_BUILD
			{ // generate debug geometry
				VERTEX_LIST  _debug_vertexList = VertexList( debugSystem_memory, 2048 );
				VERTEX_LIST * debug_vertexList = &_debug_vertexList;
				
				DEBUG_genMesh_cube(     debug_vertexList );
				DEBUG_genMesh_sphere(   debug_vertexList );
				DEBUG_genMesh_cylinder( debug_vertexList );
				DEBUG_genMesh_torus(    debug_vertexList );
				
				DEBUG_MESH       * debugMesh    = globalVar_debugSystem.debugMesh;
				RENDER_SYSTEM_3D * renderSystem = globalVar_debugSystem.renderSystem3D;
				VERTEX_BUFFER * vertexBuffer = renderSystem->vertexBuffer;
				
				for( int32 mesh_index = 1; mesh_index < debugMeshID_count; mesh_index++ ) {
					debugMesh[ mesh_index ].vertex_index += vertexBuffer->perm_baseIndex;
				}
				addPerm( vertexBuffer, debug_vertexList->vertex, debug_vertexList->vertex_count );
				
				free( debugSystem_memory, debug_vertexList );
			}
#endif // DEBUG_BUILD
			
			MOUSE_STATE    * mouse    = &win32_platform->mouse;
			KEYBOARD_STATE * keyboard = &win32_platform->keyboard;
			while( appState->isRunning ) {
#if DEBUG_BUILD
				globalVar_profileSystem.eventDatabase_clocks_previousFrame 							= globalVar_profileSystem.eventDatabase_clocks_currentFrame;
				globalVar_profileSystem.eventDatabase_clocks_currentFrame.frameStart 		  	= __rdtsc();
				globalVar_profileSystem.eventDatabase_clocks_currentFrame.counter_frameStart 	= win32_getPerfCount();
				globalVar_profileSystem.eventList_currentFrame->frameClockStart 					= globalVar_profileSystem.eventDatabase_clocks_currentFrame.frameStart;

				globalVar_atPos 	= getTL( client_bound ) + Vec2( 10.0f, -10.0f );
				globalVar_atPos	= getTL( client_bound ) + Vec2( 10.0f, -10.0f );
				
				updateProfile( &globalVar_profileSystem, win32_platform, *( ( int64 * )&perfFrequency ), globalVar_debugSystem.renderSystem2D, client_bound );
#endif // DEBUG_BUILD
				win32_processPendingMessages( win32_platform, window );
				
				POINT mouse_position = {};
				GetCursorPos( &mouse_position );
#if DEBUG_BUILD
				globalVar_profileSystem.eventDatabase_clocks_currentFrame.appStart = __rdtsc();
				if( globalVar_profileSystem.showProfile ) {
					ScreenToClient( window, &mouse_position );
					mouse->position = Vec2( ( flo32 )mouse_position.x, client_dim.y - 1.0f - ( flo32 )mouse_position.y );
				} else {
#endif // DEBUG_BUILD
					mouse->position = Vec2( ( flo32 )mouse_position.x, ( flo32 )mouse_position.y );
					
					if( wasPressed(  keyboard, keyCode_left  ) ) { appState->move_left     = true;  }
					if( wasReleased( keyboard, keyCode_left  ) ) { appState->move_left     = false; }
					if( wasPressed(  keyboard, keyCode_right ) ) { appState->move_right    = true;  }
					if( wasReleased( keyboard, keyCode_right ) ) { appState->move_right    = false; }
					if( wasPressed(  keyboard, keyCode_up    ) ) { appState->move_forward  = true;  }
					if( wasReleased( keyboard, keyCode_up    ) ) { appState->move_forward  = false; }
					if( wasPressed(  keyboard, keyCode_down  ) ) { appState->move_backward = true;  }
					if( wasReleased( keyboard, keyCode_down  ) ) { appState->move_backward = false; }
					
					if( wasPressed(  keyboard, keyCode_a ) ) { appState->move_left     = true;  }
					if( wasReleased( keyboard, keyCode_a ) ) { appState->move_left     = false; }
					if( wasPressed(  keyboard, keyCode_d ) ) { appState->move_right    = true;  }
					if( wasReleased( keyboard, keyCode_d ) ) { appState->move_right    = false; }
					if( wasPressed(  keyboard, keyCode_w ) ) { appState->move_forward  = true;  }
					if( wasReleased( keyboard, keyCode_w ) ) { appState->move_forward  = false; }
					if( wasPressed(  keyboard, keyCode_s ) ) { appState->move_backward = true;  }
					if( wasReleased( keyboard, keyCode_s ) ) { appState->move_backward = false; }
					
					appState->cursor_position  = mouse->position;
					appState->player_shotPower = clamp01( appState->player_shotPower + ( flo32 )mouse->wheelClicks / 100.0f );
					
					if( wasPressed(  keyboard, keyCode_h ) ) {
						appState->showControls = true;
						appState->helpTutorial_counter = 0;
					}
					if( wasReleased( keyboard, keyCode_h ) ) { appState->showControls = false; }
					
					APP_updateAndRender( appState, win32_platform );
					
					vec2 app_halfDim = appState->app_halfDim;
					POINT center = { ( LONG )app_halfDim.x, ( LONG )app_halfDim.y };
					// ClientToScreen( window, &center );
					SetCursorPos( center.x, center.y );
#if DEBUG_BUILD
				}
				globalVar_profileSystem.eventDatabase_clocks_currentFrame.renderStart = __rdtsc();
#endif // DEBUG_BUILD
				
				if( wasPressed( &win32_platform->keyboard, keyCode_escape ) ) {
					win32_endApplication( win32_platform, appState );
				}
				
				int32 sync_delayCounter = 0;
				boo32  waitForSync = true;
				while( waitForSync ) {
					GLenum syncStatus = glClientWaitSync( openGLSync[ vertexBuffer->temp_select ], GL_SYNC_FLUSH_COMMANDS_BIT, 0 );
					if( ( syncStatus != GL_ALREADY_SIGNALED ) && ( syncStatus != GL_CONDITION_SATISFIED ) ) {
						sync_delayCounter++;
					} else {
						waitForSync = false;
					}
				}
				
				// render 3D
				glDisable( GL_BLEND );
				glEnable( GL_DEPTH_TEST );
				glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
				
				renderSystem3D->camera_transform = perspective_transform * renderSystem3D->camera_transform;
				render3D( renderSystem3D, mainShader_transform, mainShader_modColor );
				
				// render 2D overlay
				glEnable( GL_BLEND );
				glDisable( GL_DEPTH_TEST );
				
				glUniformMatrix4fv( mainShader_transform, 1, GL_FALSE, screen_transform.elem );
				render2D( renderSystem2D, mainShader_modColor );
				
#if DEBUG_BUILD
				glDisable( GL_BLEND );
				globalVar_debugSystem.renderSystem3D->camera_transform = renderSystem3D->camera_transform;
				render3D( globalVar_debugSystem.renderSystem3D, mainShader_transform, mainShader_modColor );
				
				glEnable( GL_BLEND );
				glUniformMatrix4fv( mainShader_transform, 1, GL_FALSE, screen_transform.elem );
				render2D( globalVar_debugSystem.renderSystem2D, mainShader_modColor );
				
				globalVar_profileSystem.eventDatabase_clocks_currentFrame.frameEnd 			= __rdtsc();
				globalVar_profileSystem.eventDatabase_clocks_currentFrame.counter_frameEnd = win32_getPerfCount();
				
				if( wasPressed( &win32_platform->keyboard, keyCode_b ) ) {
					InvalidCodePath;
				}
#endif // DEBUG_BUILD
				
				vertexBuffer->temp_count = 0;
				openGLSync[ vertexBuffer->temp_select ] = glFenceSync( GL_SYNC_GPU_COMMANDS_COMPLETE, 0 );
				vertexBuffer->temp_select = ( vertexBuffer->temp_select + 1 ) % 3;
				
				SwapBuffers( win32_deviceContext );
			}
		}
	}
   
	return 0;
}

#endif	// STD_INCLUDE_FUNC